# ------------------------------------------------------------
# Auto generated test file `test_033_carbonpair.py`
# ------------------------------------------------------------
# source file   = NBTest_033_carbonpair.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 033
# test comment  = carbonpair
# ------------------------------------------------------------



from carbon import CarbonSimulatorUI, CarbonPair, P, __version__, __date__
print(f"Carbon v{__version__} ({__date__})")
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonSimulatorUI))


#


#

# ------------------------------------------------------------
# Test      033
# File      test_033_carbonpair.py
# Segment   CarbonPair
# ------------------------------------------------------------
def test_carbonpair():
# ------------------------------------------------------------
    
    Sim = CarbonSimulatorUI(pair=CarbonPair("ETH/USDC"))
    Sim
    
    assert (Sim.default_basetoken, Sim.default_quotetoken) == ('ETH', 'USDC')
    
    assert (Sim.tknb, Sim.tknq) == ('ETH', 'USDC')
    
    Sim.add_order("ETH", 10, 2000, 3000)["orders"]
    Sim.add_order("ETH", 10, 2000, 2500)["orders"]
    r = Sim.state()["orders"]
    assert list(r["pair"]) == ["ETHUSDC"]*4
    r
    
    r = Sim.amm_sells("ETH", 1)["trades"]
    assert list(r["p_unit"]) == ["USDC per ETH"]*3
    r
    
    Sim.state()["orders"]
    
    


# ------------------------------------------------------------
# Auto generated test file `test_036_LiquidityApprox.py`
# ------------------------------------------------------------
# source file   = NBTest_036_LiquidityApprox.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 036
# test comment  = LiquidityApprox
# ------------------------------------------------------------



from carbon import CarbonSimulatorUI, P, __version__, __date__
from math import sqrt
print(f"Carbon v{__version__} ({__date__})")
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonSimulatorUI))




Sim = CarbonSimulatorUI(verbose=False, raiseonerror=False, pair="ETH/USDC")
Sim

Sim.add_order("ETH", 10, 2000, 3000)
Sim.add_order("ETH", 10, 2500, 2500)
Sim.add_order("USDC", 10*1250, 1500, 1000)
Sim.add_order("USDC", 10*1250, 1250, 1250)
Sim

Sim.state()["orders"]

Sim.state().keys()

Sim.state()["orderuis"]

help(Sim.state()["orderuis"][0].liquidity_approx)


# ------------------------------------------------------------
# Test      036
# File      test_036_LiquidityApprox.py
# Segment   Order 0 ETH range
# ------------------------------------------------------------
def test_order_0_eth_range():
# ------------------------------------------------------------
    
    oui = Sim.state()["orderuis"][0]
    assert (oui.pmin, oui.pmax, oui.total_liquidity) == (2000.0, 3000.0, (10.0, 'ETH'))
    oui.pmin, oui.pmax, oui.total_liquidity
    
    r = oui.liquidity_approx(500, 600, "ETH", asperc=True)
    assert r == 0
    r
    
    r = oui.liquidity_approx(3000, 3020, "ETH", asperc=True)
    assert r == 0
    r
    
    r = oui.liquidity_approx(2000, 3000, "ETH", asperc=True)
    assert r == 1.
    r
    
    r = oui.liquidity_approx(3000, 2000, "ETH", asperc=True)
    assert r == 1
    r
    
    r = oui.liquidity_approx(2000, 2500, "ETH", asperc=True)
    assert r == 0.5
    r
    
    r = oui.liquidity_approx(2500, 3000, "ETH", asperc=True)
    assert r == 0.5
    r
    
    r = oui.liquidity_approx(2250, 2750, "ETH", asperc=True)
    assert r == 0.5
    r
    
    
    r = oui.liquidity_approx(2000, 3000, "ETH", asperc=False)
    assert r == 10.
    r
    
    r = oui.liquidity_approx(2000, 2500, "ETH", asperc=False)
    assert r == 5.
    r
    
    r = oui.liquidity_approx(2000, 2500, asperc=False)
    assert r == 5.
    r
    
    r = oui.liquidity_approx(2000, 3000, "USDC", asperc=False)
    assert int(r) == int(10*sqrt(2000*3000))
    r, sqrt(2000*3000)
    
    r = oui.liquidity_approx(2250, 2750, "USDC", asperc=False)
    assert int(r) == int(10*sqrt(2250*2750)*0.5)
    r, sqrt(2250*2750)*0.5
    

# ------------------------------------------------------------
# Test      036
# File      test_036_LiquidityApprox.py
# Segment   Order 1 ETH point
# ------------------------------------------------------------
def test_order_1_eth_point():
# ------------------------------------------------------------
    
    oui = Sim.state()["orderuis"][2]
    oui.pmin, oui.pmax, oui.total_liquidity
    
    r = oui.liquidity_approx(2000, 2499, "ETH", asperc=True)
    assert r == 0
    r
    
    r = oui.liquidity_approx(2501, 3020, "ETH", asperc=True)
    assert r == 0
    r
    
    r = oui.liquidity_approx(2499, 2501, "ETH", asperc=True)
    assert r == 1
    r
    
    r = oui.liquidity_approx(2500, 2501, "ETH", asperc=True)
    assert r == 1
    r
    
    r = oui.liquidity_approx(2499, 2500, "ETH", asperc=True)
    assert r == 0
    r
    
    r = oui.liquidity_approx(2500, 2500, "ETH", asperc=True)
    assert r == 1
    r
    
    
    r = oui.liquidity_approx(2499, 2501, "ETH", asperc=False)
    assert r == 10
    r
    
    r = oui.liquidity_approx(2499, 2501, "USDC", asperc=False)
    assert r == 25000
    r
    
    r = oui.liquidity_approx(2500, 2501, "USDC", asperc=False)
    assert r == 25000
    r
    
    r = oui.liquidity_approx(2500, 2500, "USDC", asperc=False)
    assert r == 25000
    r
    
    r = oui.liquidity_approx(2000, 3000, "USDC", asperc=False)
    assert r == 25000
    r
    

# ------------------------------------------------------------
# Test      036
# File      test_036_LiquidityApprox.py
# Segment   Order 2 USDC range
# ------------------------------------------------------------
def test_order_2_usdc_range():
# ------------------------------------------------------------
    
    oui = Sim.state()["orderuis"][4]
    oui.pmin, oui.pmax, oui.total_liquidity
    
    r = oui.liquidity_approx(500, 600, "USDC", asperc=False)
    assert r == 0
    r
    
    r = oui.liquidity_approx(1600, 2000, "USDC", asperc=False)
    assert r == 0
    r
    
    r = oui.liquidity_approx(1000, 1500, "USDC", asperc=False)
    assert r == 12500
    r
    
    r = oui.liquidity_approx(1000, 1100, "USDC", asperc=False)
    assert r == 2500
    r
    
    r = oui.liquidity_approx(1400, 1500, "USDC", asperc=False)
    assert r == 2500
    r
    
    r = oui.liquidity_approx(1000, 1100, "ETH", asperc=False)
    assert r == 2500/sqrt(1000*1100)
    r, 2500/sqrt(1000*1100)
    
    r = oui.liquidity_approx(1400, 1500, "ETH", asperc=False)
    assert int(r) == int(2500/sqrt(1400*1500))
    r, 2500/sqrt(1400*1500)
    

# ------------------------------------------------------------
# Test      036
# File      test_036_LiquidityApprox.py
# Segment   Order 3 USDC point
# ------------------------------------------------------------
def test_order_3_usdc_point():
# ------------------------------------------------------------
    
    oui = Sim.state()["orderuis"][6]
    oui.pmin, oui.pmax, oui.total_liquidity
    
    r = oui.liquidity_approx(1249, 1251, "USDC", asperc=False)
    assert r == 12500
    r
    
    r = oui.liquidity_approx(1000, 2000, "USDC", asperc=False)
    assert r == 12500
    r
    
    r = oui.liquidity_approx(1249, 1250, "USDC", asperc=False)
    assert r == 0
    r
    
    r = oui.liquidity_approx(1250, 1251, "USDC", asperc=False)
    assert r == 12500
    r
    
    r = oui.liquidity_approx(1249, 1251, "ETH", asperc=False)
    assert r == 10
    r
    
    r = oui.liquidity_approx(1000, 2000, "ETH", asperc=False)
    assert r == 10
    r
    
    r = oui.liquidity_approx(1249, 1250, "ETH", asperc=False)
    assert r == 0
    r
    
    r = oui.liquidity_approx(1250, 1251, "ETH", asperc=False)
    assert r == 10
    r
    
    


"""Alpha Router restricts the exact algo calculation to specifically n threshold number of orders."""
import itertools
from .exact_router_x0y0n import ExactRouterX0Y0N
from .base_router import *
import pandas as pd
from typing import Any
import numpy as np


@dataclass
class AlphaRouter(BaseRouter):
    """
    Alpha method - Restricts the exact algo calculation to specifically n threshold number of orders.
    """

    use_positions_matchlevel: List[int] = None
    exact_router: ExactRouterX0Y0N = ExactRouterX0Y0N()

    def get_geoprice(self, subject: int) -> DecFloatInt:
        """
        Determines the geometric price of the subject curve.
        """
        order = self.orders[subject]
        x0 = Decimal(str(order.x0))
        y0 = Decimal(str(order.y0))

        if self.use_floor_division:
            return x0 // y0
        else:
            return x0 / y0

    def amt_by_target(
            self, subject: int, dx: DecFloatInt, position_subset: List[int] = None
    ) -> DecFloatInt:
        """
        An alternative trade by target function that performs a hypothetical
        swap against a subject order not constrained by available liquidity.

        :dx:         the required target amount
        :subject:   the "subject" order index
        """
        order = self.orders[subject]
        S = Decimal(str(order.S))
        B = Decimal(str(order.B))
        y = Decimal(str(order.y))
        y_int = Decimal(str(order.y_int))

        if self.use_floor_division:
            return (
                    dx * y_int ** 2 // ((S * y + B * y_int) * (S * y + B * y_int - S * dx))
            )
        else:
            return (
                    dx * y_int ** 2 / ((S * y + B * y_int) * (S * y + B * y_int - S * dx))
            )

    def match_by_target(
            self,
            x: DecFloatInt,
            is_by_target: bool = True,
            check_sufficient_liquidity: bool = True,
            threshold_orders: int = None,
    ) -> List[Action]:
        """
        The match by target function specfic to the Alpha router that:

        a1.) Run the inputAmount
             through all orders of appropriate pairs and return the hypothetical
             outputAmount for that order. Ignores whether or not there is
             sufficient liquidity associated with these orders.

        a2.) Sort the orders by maximum return (i.e. best price)

        a3.) Take the top n orders which **combined** have available liquidity
             greater than the inputAmount. I.e. cumulatively sum the available
             liquidity from the ranked orders and take the minimum number of
             orders required to fill the trade.

        a4.) Restrict the exact algo calculation to specifically these n orders
             to return the optimal distribution for the full inputAmount
             evaluated within the threshold number of orders.
        """

        if threshold_orders == None:
            threshold_orders = 10

        # (step 1.)

        hypothetical_output_amts = {
            i: self.amt_by_target(dx=x, subject=i) for i in self.indexes
        }

        # (step 2.)
        # Order the amounts
        ordered_amts = {
            j: hypothetical_output_amts[j]
            for j in sorted(
                self.indexes, key=lambda i: hypothetical_output_amts[i], reverse=True
            )
        }

        # (step 3.)
        # Get the available liquidity and return the minimum list of orders with
        # sufficient liquidity to fulfill the inputAmount
        ordered_associated_liquidity = [self.orders[i].y for i in ordered_amts.keys()]

        results = pd.DataFrame(
            [
                hypothetical_output_amts.keys(),
                hypothetical_output_amts.values(),
                ordered_associated_liquidity,
            ],
            index=[
                "indexes",
                "hypothetical_output_amts",
                "ordered_associated_liquidity",
            ],
        )
        results = results.T.copy()
        results.sort_values(
            by=["hypothetical_output_amts", 'indexes'], ascending=[False, True], inplace=True
        )

        results.loc[:, 'sliding_index'] = [np.array(win.values.tolist(), dtype=object) for win in
                                           results.indexes.rolling(threshold_orders, min_periods=threshold_orders)]
        results.loc[:, 'sliding_available_value'] = results.ordered_associated_liquidity.rolling(threshold_orders,
                                                                                                 min_periods=threshold_orders).sum()

        results.fillna(0, inplace=True)
        results.reset_index(inplace=True, drop=True)
        # print(tabulate(results,headers=list(results.columns)))
        top_n_threshold_orders = \
        [results.sliding_index[i] for i in results.index if results.sliding_available_value[i] >= abs(x)][0]

        # (step 4.)
        # Constrain the exact router to just the top n threshold orders and match
        self.exact_router.orders = self.orders
        use_positions_matchlevel = top_n_threshold_orders  # top_n(threshold_orders)
        return self.exact_router.match(
            x=x,
            is_by_target=is_by_target,
            check_sufficient_liquidity=check_sufficient_liquidity,
            threshold_orders=threshold_orders,
            use_positions_matchlevel=use_positions_matchlevel,
        )

    def amt_by_src(
            self, subject: int, dx: DecFloatInt, position_subset: List[int] = None
    ) -> DecFloatInt:
        """
        An alternative trade by source function that performs a hypothetical
        swap against a subject order not constrained by available liquidity.

        :x:         the required source amount
        :subject:   the "subject" order index
        """
        order = self.orders[subject]
        S = Decimal(str(order.S))
        B = Decimal(str(order.B))
        y = Decimal(str(order.y))
        y_int = Decimal(str(order.y_int))

        if self.use_floor_division:
            return (
                    dx
                    * (S * y + B * y_int) ** 2
                    // (S * dx * (S * y + B * y_int) + y_int ** 2)
            )
        else:
            return (
                    dx
                    * (S * y + B * y_int) ** 2
                    / (S * dx * (S * y + B * y_int) + y_int ** 2)
            )

    def match_by_src(
            self,
            x: DecFloatInt,
            is_by_target: bool = False,
            check_sufficient_liquidity: bool = True,
            threshold_orders: int = None,
    ) -> List[Action]:
        """
        The match by target function specfic to the Alpha router that:

        a1.) Run the inputAmount
             through all orders of appropriate pairs and return the hypothetical
             outputAmount for that order. Ignores whether or not there is
             sufficient liquidity associated with these orders.

        a2.) Sort the orders by maximum return (i.e. best price)

        a3.) Take the top n orders which **combined** have available liquidity
             greater than the inputAmount. I.e. cumulatively sum the available
             liquidity from the ranked orders and take the minimum number of
             orders required to fill the trade.

        a4.) Restrict the exact algo calculation to specifically these n orders
             to return the optimal distribution for the full inputAmount
             evaluated within the threshold number of orders.
        """
        # (step 1.)
        # print(x)
        hypothetical_output_amts = {i: self.amt_by_src(dx=x, subject=i) for i in self.indexes}
        # print(hypothetical_output_amts)

        # (step 3.)
        # Get the available liquidity and return the minimum list of orders with
        # sufficient liquidity to fulfill the inputAmount
        associated_liquidity = [self.orders[i].y for i in self.indexes]

        # ordered_associated_liquidity = [self.orders[i].y for i in ordered_amts.keys()]
        # print(ordered_associated_liquidity)

        amounts = []
        effective_prices = []
        available_value = []
        for k, v in hypothetical_output_amts.items():
            if v > self.orders[k].y:
                price = self.get_geoprice(k)
                amount = self.orders[k].y
                amounts += [amount]
                effective_prices += [price]
                available_value += [price * amount]
            else:
                amounts += [v]
                price = x / v
                effective_prices += [price]
                available_value += [price * v]

        results = pd.DataFrame(
            [
                hypothetical_output_amts.keys(),
                hypothetical_output_amts.values(),
                associated_liquidity,
                amounts,
                effective_prices,
                available_value,
            ],
            index=[
                "indexes",
                "hypothetical_output_amts",
                "associated_liquidity",
                "amount",
                "effective_prices",
                "available_value",
            ],
        )
        results = results.T.copy()
        results.sort_values(
            by=["hypothetical_output_amts", 'indexes'], ascending=[False, True], inplace=True
        )

        results.loc[:, 'sliding_index'] = [np.array(win.values.tolist(), dtype=object) for win in
                                           results.indexes.rolling(threshold_orders, min_periods=threshold_orders)]
        results.loc[:, 'sliding_available_value'] = results.available_value.rolling(threshold_orders,
                                                                                    min_periods=threshold_orders).sum()

        results.fillna(0, inplace=True)
        results.reset_index(inplace=True, drop=True)
        # print(tabulate(results,headers=list(results.columns)))\
        top_n_threshold_orders = \
        [results.sliding_index[i] for i in results.index if results.sliding_available_value[i] >= abs(x)][0]

        # (step 4.)
        # Constrain the exact router to just the top n threshold orders and match
        self.exact_router.orders = self.orders
        self.use_positions_matchlevel = top_n_threshold_orders  # top_n(threshold_orders)
        return self.match(
            x=x,
            is_by_target=is_by_target,
            check_sufficient_liquidity=check_sufficient_liquidity,
            threshold_orders=threshold_orders,
        )

    def match(
            self,
            x: DecFloatInt,
            is_by_target: bool = True,
            completed_trade: bool = False,
            trade: Callable = None,
            cmp: Callable = None,
            check_sufficient_liquidity: bool = True,
            threshold_orders: int = None,
    ) -> List[Action]:
        """
        Main algorithm to handle matching a trade amount against the curves/orders.
        """
        return self.exact_router.match(
            x=x,
            is_by_target=is_by_target,
            check_sufficient_liquidity=check_sufficient_liquidity,
            threshold_orders=threshold_orders,
            use_positions_matchlevel=self.use_positions_matchlevel,
        )



"""
represents a Carbon token pair; mostly helps with not getting confused on price quotes

(c) Copyright Bprotocol foundation 2022. 
Licensed under MIT

VERSION HISTORY
v2.0 -- changed constructor to allow for slashpair string
v2.0.1 -- convert_price
v2.0.2 -- allow constructor to take CarbonPair
"""
__version__ = "2.0.2"
__date__ = "11/Dec/2022"

from dataclasses import dataclass


@dataclass
class CarbonPair:
    """
    static information about a carbon token pair

    :slashpair:     the pair string in TKNB/TKNQ form, eg "ETH/USDC"; 
                    can also be a CarbonPair instance
    :tknb:          the base token (risk token) of the pair, eg ETH*
    :tknq:          the quote token (numeraire token) of the pair, eg USDC*

    * the differentiation between numeraire and risk tokens matter only for price quotes:
    in a given pair, _all_ prices will be quote as tknn per tknr, eg USDC per ETH

    see also https://www.investopedia.com/terms/i/isocurrencycode.asp for ISO currency code
    """
    __VERSION__ = __version__
    __DATE__    = __date__
    
    slashpair: any = None
    tknb: str = None 
    tknq: str = None

    def __post_init__(self):

        if self.slashpair:
            #print("[CarbonPair] slashpair has been provided", self.slashpair, self.tknb, self.tknq)
            if self.tknb or self.tknq:
                if self.tknb:
                    raise ValueError(
                        f"Parameters are pair, tknb, tknq; did you mean `tknb='{self.slashpair}', tknq='{self.tknb}'` ?",
                        self.slashpair, self.tknb, self.tknq)
                raise ValueError("Must not provide both pair and tknb, tknq", self.slashpair, self.tknb, self.tknq)
            
            if isinstance(self.slashpair, self.__class__):
                #print("[CarbonPair] creating pair from CarbonPair", self.__class__, self.slashpair.__class__, self.slashpair.slashpair)
                self.slashpair = self.slashpair.slashpair
            else:
                pass
                #print("[CarbonPair] creating pair from string", self.__class__, self.slashpair)

            self.slashpair = self.slashpair.upper()

            try:
                self.tknb, self.tknq = self.slashpair.split("/")
            except:
                raise ValueError("Illegal slashpair", self.slashpair)
            #print("[CarbonPair] {self.slashpair} -> tknb={self.tknb}, tknq={self.tknq}")        
        
        else:
            #print("[CarbonPair] tknb, tknq has hopefully been provided", self.tknb, self.tknq)
            if self.tknb is None or self.tknq is None:
                raise ValueError("If pair is None must provide tknb, tknq", self.slashpair, self.tknb, self.tknq)            
            self.tknb = self.tknb.upper()
            self.tknq = self.tknq.upper()
            self.slashpair = f"{self.tknb}/{self.tknq}"

    def __repr__(self):
        return f"P('{self.slashpair}')"
        return f"{self.__class__.__name__}('{self.slashpair}')"

    @classmethod
    def from_isopair_and_tkn(cls, isopair, tkn=None):
        """
        creates a class instance from and iso pair string and a token

        :isopair:   the pair, in ISO format (BASQUO)
        :tkn:       any token that is part of the pair*

        *tkn is only used to separate the two tokens in the pair string; it is is None,
        and the lenght of the pair string is 3 or 8, then it is split in the middle; if
        None and odd an exception is raised
        """
        isopair = isopair.upper()
        if tkn is None:
            if len(isopair) == 6:
                tknb, tknq = isopair[:3], isopair[3:]
            elif len(isopair) == 8:
                tknb, tknq = isopair[:4], isopair[4:]
            else:
                raise ValueError(
                    "If tkn not given, length of isopair must be 6 or 8",
                    isopair,
                    len(isopair),
                    tkn,
                )
        else:
            tkn2 = isopair.replace(tkn, ".")
            if tkn2 == isopair:
                raise ValueError(
                    "Invalid token specification (tkn not part of isopair)",
                    isopair,
                    tkn,
                )
            if tkn2[0] == ".":
                tknb = tkn
                tknq = tkn2[1:]
            elif tkn2[-1] == ".":
                tknb = tkn2[:-1]
                tknq = tkn
            else:
                raise ValueError(
                    "Invalid token specification (tkn inside isopair)", isopair, tkn
                )

        return cls(tknb=tknb, tknq=tknq)

    @classmethod
    def from_slashpair(cls, slashpair):
        """
        creates a pair from a slashstring (tknb/tknq, eg "ETH/USDC")

        :slashpair:     the pair, in the format tknb/tknq, eg "ETH/USDC"
        """
        slashpair = slashpair.upper()
        tokens = slashpair.split("/")
        if not len(tokens) == 2:
            raise ValueError("slashpair must be of form TKNB/TKNQ", slashpair, tokens)
        tknb = tokens[0].strip()
        tknq = tokens[1].strip()
        return cls(f"{tknb}/{tknq}")
        

    @classmethod
    def create(cls, arg1=None, arg2=None):
        """
        alternative constructor, allowing various paramerisations

        EXAMPLES

        p = Carbonpair.create("ETH/USDC")   # Method1: from slashpair
        Carbonpair.create(p)                # Method 2: from other CarbonPair
        Carbonpair.create("ETH", "USDC")    # Method 3: quote token, base token
        Carbonpair.create()                 # None
        """

        if arg1 is None and arg2 is None:
            return None

        if isinstance(arg1, cls):
            if not arg2 is None:
                raise ValueError("Second argument must be None if arg1 is pair", arg1, arg2)
            return arg1
        if arg2 is None:
            return cls(arg1)
        return cls(tknb=arg1, tknq=arg2)

    
    @property
    def price_convention(self):
        """
        returns the price convention of the pair, eg 'USDC per ETH'
        """
        return f"{self.tknq} per {self.tknb}"

    # @property
    # def pair_id(self):
    #     """
    #     returns the canonic Carbon id of the pair

    #     The Carbon ID of the pair is the concatenation of the token names
    #     in alphabetical order, in upper case, separated by colon.
    #     A carbon id does not imply any specific quote conventions
    #     """
    #     if self.tknb < self.tknq:
    #         return f"{self.tknb}:{self.tknq}"
    #     return f"{self.tknq}:{self.tknb}"

    # @property
    # def pair_id_is_reversed(self):
    #     """
    #     returns True iff Carbon id is reversed compared to iso

    #     if the carbon id is TKN1:TKN2 then ISO id of TKN1TKN2 will yield False, and TKN2TKN1 True
    #     """
    #     return self.tknb > self.tknq


    @property
    def pair_iso(self):
        """
        returns the name of the pair in iso format*, ie BASQUO

        *see https://www.investopedia.com/terms/i/isocurrencycode.asp
        """
        return f"{self.tknb}{self.tknq}"

    @property
    def pair_slash(self):
        """convenience access for slashpair"""
        return self.slashpair

    @property
    def reverse(self):
        """
        returns the pair with base and quote token reversed
        """
        return self.__class__(tknb=self.tknq, tknq=self.tknb)

    @property
    def basetoken(self):
        """alias for tknb"""
        return self.tknb 
    
    def has_basetoken(self, tkn):
        """
        returns True if `tkn` is the base token of the pair, otherwise False
        """
        return tkn == self.basetoken

    @property
    def quotetoken(self):
        """alias for tknq"""
        return self.tknq 
    
    def has_quotetoken(self, tkn):
        """
        returns True if `tkn` is the quote token of the pair, otherwise False
        """
        return tkn == self.quotetoken

    def has_token(self, tkn):
        """
        returns True if `tkn` is one of the tokens of the pair, otherwise False
        """
        if self.has_quotetoken(tkn):
            return True
        return self.has_basetoken(tkn)

    def other(self, tkn):
        """
        returns the other token, ie the one that is not `tkn`
        """
        if self.has_basetoken(tkn):
            return self.tknq
        if self.has_quotetoken(tkn):
            return self.tknb
        return None

    def convert(self, amtfrom, tknfrom, tknto, price):
        """
        converts one token amount into the other, with price according to curve convention

        :amtfrom:   the amount of tknfrom
        :tknfrom:   source token
        :tknto:     target token; can be the same as source token, in which case price is unity
        :price:     the price, in the price_convention of the pair
        :returns:   the amount of target token to be returned, or None if one of the tokens not in pair
        """
        if not self.has_token(tknfrom):
            return None
        if not self.has_token(tknto):
            return None
        if tknfrom == tknto:
            return amtfrom

        ismult = self.has_basetoken(tknfrom)
        if ismult:
            return amtfrom * price
        return amtfrom / price

    def price(self, price0, tknb0, tknq0):
        """
        the normalized price, according to convention

        :price0:    the raw price, as tknq0 per tknb0
        :tknb0:     the base token of price0
        :tknq0:     the quote token of price0
        :returns:   the amount of tarket token to be returned
        """
        if not self.has_token(tknb0):
            return None
        if not self.has_token(tknq0):
            return None
        if tknb0 == tknq0:
            return None

        isreverse = self.has_basetoken(tknq0)
        if isreverse:
            return 1.0 / price0
        return price0

    def convert_price(self, price, tknq):
        """
        converts a price expressed in the tknq numeraire in the numeraire of the pair

        :price:     the price to be converted, with numeraire being `tknq`
        :tknq:      the numeraire token in which the `price` is expressed
        :returns:   the price expressed in the numeraire conventions of the pair;
                    returns None if tknq not part of this pair or price is None
        """
        if price is None:
            return None
        if not self.has_token(tknq):
            return None
        if self.has_quotetoken(tknq):
            return price
        return 1 / price

    BUY = "buy"
    SELL = "sell"
    MAKERBUY = "buy"  # the MAKER is the person who placed the order
    MAKERSELL = (
        "sell"  # they own the optionality, ie they will only buy if in the money
    )
    TAKERBUY = "sell"  # the TAKER is the counterparty to this order; they have no
    TAKERSELL = (
        "buy"  # optionality here, so they'll execute iff it is beneficial for the MAKER
    )

    def limit_is_met(
        self, tkn, limit_price, buysell, current_price, reverse=False, asphrase=False
    ):
        """
        checks whether a limit order has been met

        :tkn:               the token to be bought or sold
        :limit_price:       the price at which the token is to be bought or sold; quoted
                            in the price convention of this pair
        :buysell:           whether the limit order is for buying or selling `tkn`; this is
                            from the point of view of the person who PLACED the order; use
                            `reverse=True` to do it from the point of view of the person
                            who ACCEPTED the orders; use the constants BUY and SELL
        :current_price:     the price which is tested against the limit order
        :asphrase:          returns the result as an explanatory phrase instead of bool
        :returns:           True if the limit order will lead to execution, false if not
        """
        if not self.has_token(tkn):
            print(f"[limit_is_met] invalid token {tkn} for pair {self.pair_iso}")
            return None

        if asphrase:
            limit_is_met = self.limit_is_met(
                tkn, limit_price, buysell, current_price, reverse
            )
            action = f"{buysell}s" if limit_is_met else f"does not {buysell}"
            result = f"Order placer {action} {tkn} at {current_price} {self.price_convention} (limit={limit_price})"
            return result

        # if the prices coincide, both buy and sell orders will be executed
        if current_price == limit_price:
            return True

        reverse = self.has_quotetoken(tkn)

        if buysell == self.BUY:
            # the order placer BUYs at limit or better
            # print(f"[limit_is_met] BUY cp={current_price} lp={limit_price} cp<=lp:{current_price<=limit_price}", )
            if not reverse:
                return (
                    current_price < limit_price
                )  # order placer only BUYS at limit or BELOW
            else:
                return current_price > limit_price  # ...reverse

        elif buysell == self.SELL:
            # the order placer SELLs at limit or better
            # print(f"[limit_is_met] SELL cp={current_price} lp={limit_price} cp<=lp:{current_price<=limit_price}", )
            if not reverse:
                return (
                    current_price > limit_price
                )  # order placer only SELLS at limit or ABOVE
            else:
                return current_price < limit_price  # ...reverse
        else:
            print(f"[limit_is_met] invalid choice for buysell [{buysell}]")



# ------------------------------------------------------------
# Auto generated test file `test_048_PartialFullfilment.py`
# ------------------------------------------------------------
# source file   = NBTest_048_PartialFullfilment.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 048
# test comment  = PartialFullfilment
# ------------------------------------------------------------



import pandas as pd
from decimal import Decimal

from carbon import CarbonSimulatorUI, __version__, __date__
print(f"Carbon Version v{__version__} ({__date__})", )


#


# ------------------------------------------------------------
# Test      048
# File      test_048_PartialFullfilment.py
# Segment   Init
# ------------------------------------------------------------
def test_init():
# ------------------------------------------------------------
    
    # Initialize a fast simulator
    
    Sim = CarbonSimulatorUI(pair="ETH/USDC", verbose=False, raiseonerror=True)
    
    Sim.add_strategy("ETH", 10, 2000, 2000, 10000, 1000, 1000)
    Sim.add_strategy("ETH", 10, 2010, 2010, 10000, 1010, 1010)
    Sim.add_strategy("ETH", 10, 2020, 2020, 10000, 1020, 1020)
    Sim.add_strategy("ETH", 10, 2030, 2030, 10000, 1030, 1030)
    Sim.add_strategy("ETH", 10, 2040, 2040, 10000, 1040, 1040)
    Sim.state()["orders"]
    
    # ### Route by Source
    
    # #### AMM buys ETH
    
    # Start with an acceptable amount and support_partial set to False
    
    results = Sim.amm_buys('ETH',10, execute=False, support_partial=False)['trades']  # route_trade_by_source
    results
    
    # Then try an unacceptable amount with support_partial still set to False
    #
    # Should FAIL
    
    try:
        results = Sim.amm_buys('ETH',1000000, execute=False, support_partial=False)['trades']  # route_trade_by_source
    except AssertionError as e:
        print('Should FAIL with insufficient liquidity...')
        assert(True)
        print(e)
    
    # Then set support_partial = True and pass an acceptable amount
    
    results = Sim.amm_buys('ETH',10, execute=False, support_partial=True)['trades']  # route_trade_by_source
    results
    
    # Then set support_partial = True and pass an unacceptable amount
    #
    # Should NOT FAIL
    
    results = Sim.amm_buys('ETH',10000000, execute=False, support_partial=True)['trades']  # route_trade_by_source
    assert(f"{results[results.uid=='2']['amt1'].values[0]:.0f}" == "50000")
    results
    
    # Now test the other combinations
    
    # #### AMM buys USDC
    
    try:
        results = Sim.amm_buys('USDC',1000000, execute=False, support_partial=False)['trades']  # route_trade_by_source
    except AssertionError as e:
        print('Should FAIL with insufficient liquidity...')
        assert(True)
        print(e)
    
    results = Sim.amm_buys('USDC',10000000, execute=False, support_partial=True)['trades']  # route_trade_by_source
    assert(f"{results[results.uid=='3']['amt1'].values[0]:.0f}" == "50")
    results
    
    # ### Route by Target
    
    # #### AMM sells ETH
    
    try:
        results = Sim.amm_sells('ETH',1000000, execute=False, support_partial=False)['trades']  # route_trade_by_source
    except AssertionError as e:
        print('Should FAIL with insufficient liquidity...')
        assert(True)
        print(e)
    
    results = Sim.amm_sells('ETH',10000000, execute=False, support_partial=True)['trades']  # route_trade_by_source
    assert(f"{results[results.uid=='4']['amt1'].values[0]:.0f}" == "50")
    results
    
    # #### AMM sells USDC
    
    try:
        results = Sim.amm_sells('USDC',1000000, execute=False, support_partial=False)['trades']  # route_trade_by_source
    except AssertionError as e:
        print('Should FAIL with insufficient liquidity...')
        assert(True)
        print(e)
    
    results = Sim.amm_sells('USDC',10000000, execute=False, support_partial=True)['trades']  # route_trade_by_source
    assert(f"{results[results.uid=='5']['amt1'].values[0]:.0f}" == "50000")
    results


# ------------------------------------------------------------
# Auto generated test file `test_037_Match.py`
# ------------------------------------------------------------
# source file   = NBTest_037_Match.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 037
# test comment  = Match
# ------------------------------------------------------------



from carbon import CarbonSimulatorUI, P, __version__, __date__
print(f"Carbon v{__version__} ({__date__})")
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonSimulatorUI))







Sim = CarbonSimulatorUI(raiseonerror=False, pair="ETH/USDC")
assert(Sim._mm==CarbonSimulatorUI.MATCH_EXACT)
Sim

SimE = CarbonSimulatorUI(raiseonerror=False, pair="ETH/USDC", matching_method=CarbonSimulatorUI.MATCH_EXACT)
assert(SimE._mm==CarbonSimulatorUI.MATCH_EXACT)
SimE

SimA = CarbonSimulatorUI(raiseonerror=False, pair="ETH/USDC", matching_method=CarbonSimulatorUI.MATCH_ALPHA)
assert(SimA._mm==CarbonSimulatorUI.MATCH_ALPHA)
SimA

#SimF = CarbonSimulatorUI(raiseonerror=False, pair="ETH/USDC", matching_method=CarbonSimulatorUI.MATCH_FAST)
SimF = CarbonSimulatorUI(raiseonerror=False, pair="ETH/USDC")
#assert(SimA._mm==CarbonSimulatorUI.MATCH_FAST)
SimF

SIMS = [Sim, SimE, SimA, SimF]



ORDERS = [
    ("ETH", 1, 2000+10*i, 3000)
    for i in range(20)   
]
ORDERS[:2]

for sim in SIMS:
    for o in ORDERS:
        sim.add_order(*o)
    print(sim, len(sim.state()["orders"]), len(ORDERS))
    assert(len(sim.state()["orders"])==2*len(ORDERS))



Sim.amm_sells("ETH", 1, execute=False)["trades"]

SimA.amm_sells("ETH", 1, execute=False)["trades"]




"""
Carbon helper module - standard Carbon imports for sims

USAGE

    from carbon.helper.stdimports import *

This will print version numbers of key Carbon modules
"""
from math import sqrt, exp, log
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

from carbon import CarbonSimulatorUI, CarbonOrderUI, P, __version__, __date__

print(f"[stdimports] Carbon v{__version__} ({__date__})")
print( "[stdimports] {0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonSimulatorUI))
print( "[stdimports] {0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonOrderUI))



from carbon import CarbonSimulatorUI


def test_inverted_range_positions_1():
    """
    Derived from `passed_tests` notebook test-13
    """
    Sim = CarbonSimulatorUI(verbose=True)
    Sim.add_order("ETH", 10, 2000, 3000, pair="ETH/USDC")
    result = Sim.add_order("ETH", 10, 3000, 2000, pair="ETH/USDC")
    assert result["success"] is True
    assert list(Sim.state()["orders"]["p_end"].values) == [3000.0, None, 3000.0, None]


def test_inverted_range_positions_2():
    """
    Derived from `passed_tests` notebook test-13
    """
    Sim = CarbonSimulatorUI(verbose=True)
    result = Sim.add_strategy("BAT", 10, 12.0, 10, 0, 5, 7.5, pair="BAT/MAN")
    assert result["success"] is True
    assert list(Sim.state()["orders"]["p_end"].values) == [12.0, 5.0]



"""
wrapper UI class for the Carbon simulation

(c) Copyright Bprotocol foundation 2022. 
Licensed under MIT

VERSION HISTORY
v2.0 - require slashpair notation (breaking change); exclude_future, constants for matching method
v2.2 - curve disabling, single orders
v2.2.1 - CarbonOrderUI linked
v2.3 - added fast router
v2.4 - limit orders, changed trade_action -> match_by in _trade
"""
__version__ = "2.4"
__date__ = "12/Jan/2023"

import itertools
from typing import Callable, Any, Tuple, Dict, List

from tabulate import tabulate

from ..order import Order
from ..pair import CarbonPair
from ..carbon_order_ui import CarbonOrderUI
from ..routers import ExactRouterX0Y0N, AlphaRouter, FastRouter
from decimal import Decimal
import pandas as pd
import numpy as np



class CarbonSimulatorUI:
    """
    Providing the user interface for the Carbon simulation

    :verbose:           if True (default), the simulation logger. infos out info along the way
    :pair:              if given, Simulation methods uses this pair as default; other pairs
                        can still be used, but need to be explicitly mentioned; pairs must be provided 
                        either as CarbonPair, or in the slash format "ETH/USDC"
    :raiseonerror:      if False (default), errors are caught and returned to the result dict
    :matching_method:   if MATCH_EXCACT (default), the exact matching algorithm is used; otherwise 
                        choices are MATCH_FAST and MATCH_ALPHA
    :decimals:          the number of decimals to which the output is rounded
    :exclude_future:    if True (default), excludes future and experimental features
    """

    __VERSION__ = __version__
    __DATE__    = __date__

    MATCH_EXACT = "exact"
    MATCH_FAST = "fast"
    MATCH_ALPHA = "alpha"

    def __init__(
            self,
            verbose: bool = False,
            pair: Any = None,
            raiseonerror: bool = False,
            decimals: int = 50,
            matching_method: str = MATCH_EXACT,
            exclude_future: bool = True,
    ):
        self._pos_id = itertools.count()
        self.verbose = verbose
        self.debug = False  # setting this to True enables debug output
        self.exclude_future = exclude_future

        self._carbon_pair = CarbonPair.create(pair)

        self.raiseonerror = raiseonerror
        self.numtrades = 0
        self.decimals = decimals
        self._mm = matching_method
        if matching_method == self.MATCH_ALPHA:
            self.matcher = AlphaRouter(verbose=False)
        elif matching_method == self.MATCH_EXACT:
            self.matcher = ExactRouterX0Y0N(verbose=False)
        elif matching_method == self.MATCH_FAST:
            self.matcher = FastRouter(verbose=False)
        elif matching_method == self.MATCH_FAST:
            raise ValueError("Fast router not implemented", matching_method)
        else:
            raise ValueError("Illegal value for matching_method", matching_method)

        self.orders = {}
        self.vault = {}
        self.trades = {}

        if self.verbose:
            print(
                f"[__init__] pair={pair}, verbose={verbose}, raiseonerror={raiseonerror}"
            )

    class ExcludedFutureFunctionality(ValueError):
        """
        exception raised when functionality requested is considered future, and `excludefuture` is True
        """
        pass

    def _raise_if_future_restricted(self, msg=None, *args, **kwargs):
        """
        raises ExcludedFutureFunctionality exception if self.exclude_future is True

        :msg, args, kwargs:     arguments for the exception
        """
        if self.exclude_future:
            if msg is None:
                msg = "Feature disabled (us `exclude_future = False` to enable)"
            raise self.ExcludedFutureFunctionality(msg, *args, **kwargs)

    @property
    def default_basetoken(self):
        """
        returns the default base token of the sim (only if initialised with CarbonPair)
        """
        return self._carbon_pair.tknb if self._carbon_pair else None

    tknb = default_basetoken

    @property
    def default_quotetoken(self):
        """
        returns the default quote token of the sim (only if initialised with CarbonPair)
        """
        return self._carbon_pair.tknq if self._carbon_pair else None

    tknq = default_quotetoken

    def get_carbon_pair(self, pair = None):
        """
        helper functon determining the pair from function params and class defaults

        :pair:      a slashpair-formt string determining the pair, or CarbonPair
        :returns:   CarbonPair object, or raises
        """
        if pair:
            return CarbonPair.create(pair)

        if self._carbon_pair is None:
            raise ValueError("Pair must be provided in function or simulation defaults")

        return self._carbon_pair

    @property
    def carbon_pair(self):
        """return the default CarbonPair associated with this object, or None"""
        return self._carbon_pair

    @property
    def slashpair(self):
        """return the slashpair or None"""
        return self._carbon_pair.slashpair if self._carbon_pair else None

    def price_convention(self, pair=None):
        """
        gets the price convention associated with `pair`

        :pair:      a string describing the pair (eg, "ETH/USDC"); if None (default)
                    use the price convention of the Sim object; raises an error if
                    no pair specified at all
        :returns:   a string describing the price convention (eg "USDC per ETH")
        """
        carbon_pair = self.get_carbon_pair(pair)
        return carbon_pair.price_convention

    def _add_order_sell_tkn(
            self,
            tkn: str,
            amt: Any,
            p_lo: Any,
            p_hi: Any,
            carbon_pair: CarbonPair,
            id1: int = None,
            id2: int = None,
            p_marginal: Any = None,
    ) -> int:
        """
        PRIVATE - adds a position for sale of tkn

        :tkn:           the token that is being added to the position, eg "ETH"; it is the token being sold*
        :amt:           the amount of `tkn` that is added to the position
        :p_lo:          the lower end of the range, quoted with `tkn` as numeraire; also p_b, p_end
        :p_hi:          ditto upper; also p_a, p_start
        :carbon_pair:   the token pair a CarbonPair object
        :id1/2:         the order ids of the two orders that make up the position; if not given, they are generated
        :p_marginal:    the marginal price of the position; if not given, it is calculated
        :returns:       the id of the order added
        """

        # create order tracking ids
        if id1 is None:
            id1 = self._posid
        if id2 is None:
            id2 = id1

        if amt is None:
            amt = 0

        if not (p_lo is None and p_hi is None):

            # check that they are both numbers
            if p_lo is None or p_hi is None:
                raise ValueError("p_lo, p_hi must either be both None or both numbers", p_lo, p_hi)

            # rearrange p_lo, p_hi if need be
            if not p_lo <= p_hi:
                pp = p_lo
                p_lo = p_hi
                p_hi = pp
                print("WARNING: swapped start and end of the range")

            disabled = False

        else:
            disabled = True
            p_lo = 0
            p_hi = 0

        # enter order
        order_params = {
            "tkn": tkn,                                     # the token being sold
            "_y": amt,                                       # the marginal price of the curve
            "p_low": p_lo,                                  # the lower end of the range (`tkn` numeraire); also p_b, p_end
            "p_high": p_hi,                                 # the upper end of the range (`tkn` numeraire); also p_a, p_start
            "disabled": disabled,                           # if True, order is disabled (p=0)
            "pair_name": carbon_pair.pair_iso,              # the iso name of the pair
            "pair": carbon_pair,                            # the CarbonPair object
            "id": id1,                                      # the id of the new curve generated
            "linked_to_id": id2,                            # the id to which this curve is linked (=id if single curve)
        }
        if p_marginal is not None:
            order_params["p_marginal"] = p_marginal         # the marginal price of the curve

        if p_marginal is None:
            order_params["y_int"] = amt                     # the capacity of the curve

        self.orders[id1] = Order(**order_params)
        return id1

    def add_order(
            self, tkn: str, amt: Any, p_start: Any, p_end: Any, pair: Any = None
    ) -> Dict[str, Any]:
        """
        adds a sell order for tkn

        :tkn:           the token that is being added to the position, eg "ETH"; it is the token being sold*
        :amt:           the amount of `tkn` that is added to the position
        :p_start:       the start* of the range, quoted in the currency of the pair
        :p_end:         ditto end*
        :pair:          the token pair to which the position corresponds, eg "ETHUSD"*

        *p_start, p_end are interchangeable, the code deals with sorting them correctly,
        albeit with a warning message if they were in the wrong order
        """
        return self.add_strategy(tkn, amt, p_start, p_end, pair=pair)

    def _add_replace_single_order(
            self, tkn: str, amt=None, p_start=None, p_end=None, pair=None, oid=None, lid=None
    ) -> Dict[str, Any]:
        """
        adds or modifies a single sell order for tkn (not part of public interface)

        :tkn:           the token that is being added to the position, eg "ETH"; it is the token being sold*
        :amt:           the amount of `tkn` that is added to the position
        :p_start:       the start* of the range, quoted in the currency of the pair
        :p_end:         ditto end*
        :pair:          the token pair to which the position corresponds, eg "ETHUSD"*
        :oid:           the id of the order; if None, a new id is created
        :lid:           the linked order id; if None, linked to itself**

        *p_start, p_end are interchangeable, the code deals with sorting them correctly,
        albeit with a warning message if they were in the wrong order

        **if orders are linked to themselves then the tokens the curve receives are not
        stored or recorded anywhere; they simply disappear
        """
        try:
            # this is a non-MVP feature
            self._raise_if_future_restricted()

            # validate tkn, pair and get CarbonPair object
            carbon_pair = self.get_carbon_pair(pair)
            if not carbon_pair.has_token(tkn):
                raise ValueError("Token not part of pair", tkn, carbon_pair, pair)

            # create order tracking ids
            if oid is None:
                id1 = self._posid
                #print(f"[_add_modify_single_order] generating order with id {id1}")
            else:
                id1 = oid

            if lid is None:
                id2 = id1
                #print(f"[_add_modify_single_order] setting lid to oid = {id2}")
            else:
                id2 = lid
                

            # convert the prices from the pair numeraire to the curve numeraire
            p_start_c = carbon_pair.convert_price(p_start, tkn)
            p_end_c = carbon_pair.convert_price(p_end, tkn)

            # ugly hack, but somehow we need to switch lo and hi to have the
            # boundary closer to the money always first
            order_id = self._add_order_sell_tkn(
                tkn, amt, p_end_c, p_start_c, carbon_pair, id1, id2
            )

            if self.verbose:
                print(
                    f"[add_sgl_pos] added position: tkn={tkn}, amt={amt}, p_start={p_start}, p_end={p_end}, "
                    f"pair={carbon_pair.pair_iso}, id={id1}, lid={id2}"
                )

            orders = self._to_pandas(self.orders[order_id], decimals=self.decimals)
            orderuis = {order_id: CarbonOrderUI.from_order(self.orders[order_id])}

        except Exception as e:
            if self.raiseonerror:
                raise
            return {"success": False, "error": str(e), "exception": e}
        return {"success": True, "orders": orders, "orderuis": orderuis, "id": id1, "lid": id2}

    #add_sgl_pos = add_sellorder
    #add_order = add_sellorder

    def add_strategy(
            self,
            tkn: str,
            amt_sell: Any = None,
            psell_start: Any = None,
            psell_end: Any = None,
            amt_buy: Any = None,
            pbuy_start: Any = None,
            pbuy_end: Any = None,
            pair: str = None,
            psell_marginal: Any = None,
            pbuy_marginal: Any = None,
    ) -> Dict[str, Any]:
        """
        adds two linked orders (one buy, one sell; aka a "strategy")

        :tkn:           the token that is sold in the range psell_start/_end, eg "ETH"*
        :amt_sell:      the amount of `tkn` that is available for sale in range psell_start/psell_end
        :psell_start:   start of the sell `tkn` range*, quoted in the price convention of `pair`
        :psell_end:     ditto end
        :amt_buy:       the amount of the other token that is available for selling against tkn in range pbuy_start
        :pbuy_start:    start of the of the buy `tkn` range*, quoted in the price convention of `pair`
        :pbuy_end:      ditto end
        :pair:          the token pair to which the strategy corresponds, eg "ETHUSD"
        :pbuy_marginal: the current price of the other token in the pair
        :psell_marginal: the current price of `tkn` in the pair

        *px_start, px_end are interchangeable, the code deals with sorting them, albeit issuing a warning
        message if they are in the wrong order; sell and buy is seen from the perspective of the AMM, which
        is the same as that of the strategy (liquidity) provider

        amounts that are None (default) are set to zero
        pbuy that are None are effectively set to 0 (disabled)
        psell that are None are effectively set to 1/0 (disabled)
        """
        try:

            # validate tkn, pair and get CarbonPair object
            carbon_pair = self.get_carbon_pair(pair)
            tkn2 = carbon_pair.other(tkn)
            if tkn2 is None:
                raise ValueError("Token not in pair", tkn, carbon_pair.slashpair, pair.slashpair)

            # create order tracking ids
            id1 = self._posid
            id2 = self._posid

            # convert the prices from the pair numeraire to the curve numeraire
            psell_marginal_c = carbon_pair.convert_price(psell_marginal, tkn)
            psell_start_c = carbon_pair.convert_price(psell_start, tkn)
            psell_end_c = carbon_pair.convert_price(psell_end, tkn)
            pbuy_start_c = carbon_pair.convert_price(pbuy_start, tkn2)
            pbuy_end_c = carbon_pair.convert_price(pbuy_end, tkn2)
            pbuy_marginal_c = carbon_pair.convert_price(pbuy_marginal, tkn2)

            # ugly hack but the 1/2 range boundaries are in the wrong order
            # we want the closer boundary first
            self._add_order_sell_tkn(
                tkn, amt_sell, psell_end_c, psell_start_c, carbon_pair, id1, id2, psell_marginal_c
            )
            self._add_order_sell_tkn(
                tkn2, amt_buy, pbuy_end_c, pbuy_start_c, carbon_pair, id2, id1, pbuy_marginal_c
            )

            if self.verbose:
                print(
                    f"[add_linked_pos] added tkn={tkn}, amt_sell={amt_sell}, psell_start={psell_start}, psell_marginal_c={psell_marginal_c}, "
                    f"psell_end={psell_end}, amt_buy={amt_buy}, pbuy_start={pbuy_start}, pbuy_end={pbuy_end}, pbuy_marginal_c={pbuy_marginal_c},"
                    f"pair={carbon_pair.pair_iso}"
                )

            orders = pd.concat(
                [
                    self._to_pandas(o, decimals=self.decimals)
                    for o in [self.orders[id1], self.orders[id2]]
                ]
            )
            orderuis = {
                id1: CarbonOrderUI.from_order(self.orders[id1]),
                id2: CarbonOrderUI.from_order(self.orders[id2]),
            }

        except Exception as e:
            if self.raiseonerror:
                raise
            return {"success": False, "error": str(e), "exception": e}
        return {"success": True, "orders": orders, "orderuis": orderuis}

    #add_linked_pos = add_strategy

    def delete_order(self, position_id):
        """
        deletes a positon / order or strategy

        :position_id:   the index of (one of) the orders to be deleted; this order, and all
                        orders linked to it will be removed
        :returns:       the orders removed as list, [] if not found
        """

        try:
            order = self.orders[position_id]
        except KeyError as e:
            if self.verbose:
                print(f"Position {position_id} does not exist: {e}")
            return []

        linked_position_id = order.linked_to_id
        if linked_position_id == position_id:
            linked_position_id = None

        pair = order.pair

        del self.orders[position_id]

        if linked_position_id:
            del self.orders[linked_position_id]

        if self.verbose:
            str2 = (
                f"and linked position {linked_position_id}"
                if linked_position_id
                else ""
            )
            print(f"Deleted position {position_id} {str2}[{pair}]")

        return (
            [position_id, linked_position_id] if linked_position_id else [position_id]
        )

    #delete_pos = delete_order
    delete_strategy = delete_order

    MATCH_BY_SOURCE = "match_by_source"
    MATCH_BY_TARGET = "match_by_target"

    def _trade(
            self,
            tkn: str,
            amt: Any,
            support_partial: bool,
            carbon_pair: CarbonPair,
            #trade_action: Callable = None,
            match_by: str = None,
            trade_description: str = None,
            execute: bool = True,
            limit_price: Any = None,
            limit_amt: Any = None,
            inpair: bool = True,
            use_positions: List[int] = None,
            threshold_orders: int = None,
            use_positions_matchlevel: List[int] = [],
            is_by_target: bool = False,

    ) -> Dict[str, Any]:
        """
        PRIVATE - executes a trade

        :tkn:                   the token that is being SOLD by the AMM, eg "ETH"*
        :amt:                   the amount to be traded*
        :carbon_pair:           the CarbonPair class of the token pair
        :match_by:              either MATCH_BY_SOURCE or MATCH_BY_TARGET
        :trade_description:     human-readable description of the trade, eg "amm vs trader, buy vs sell, for amt"
        :execute:               if True (default), the trade is executed; otherwise only routing is shown
        :limit_price:           the limit price of the order (this price or better from point of view
                                of the trader, not the AMM!), quoted in convention of the pair**
        :limit_amt:             the minimum amount of tokens the traders expects to obtain; quoted in units of the
                                other token, ie `carbon_pair.other(tkn)`**
        :inpair:                if True, only match within pair; if False (default), route through all available pairs
        :use_positions:         the positions to use for the trade (default: all positions)
        :threshold_orders:      the maximum number of order to be routed through using the alpha router
        :support_partial:       if True (not default), and insufficient liquidity for a trade request, a partial fullfilment is made

        *amt is always effectively a positive amount; however, if `match_by` is `MATCH_BY_TARGET` then it must
        be provided as a negative number, and if `MATCH_BY_SOURCE` as positive number

        **limit_price and limit_amt are redundant; either one of them can be given, or none, but not both
        """
        trades, orders = None, None
        is_partial = False

        if match_by == self.MATCH_BY_TARGET:
            trade_action = self.matcher.match_by_target
        elif match_by == self.MATCH_BY_SOURCE:
            trade_action = self.matcher.match_by_src
        else:
            raise ValueError("match_by must be self.MATCH_BY_TARGET or self.MATCH_BY_SOURCE", match_by)

        if not limit_amt is None:
            if not limit_price is None:
                raise ValueError("Not both limit_amt and limit_price can be given", limit_amt, limit_price)
            
            # amt is expressed in the source token if match_by == MATCH_BY_SOURCE, otherwise target token
            # rawprice is expressed as tkn per other token if MATCH_BY_TARGET, otherwise reverse
            # we use carbon_pair.convert_price to convert it into the proper price convention
            rawprice = abs(float(amt) / float(limit_amt))
            limit_price = carbon_pair.convert_price(
                rawprice, tknq = tkn if match_by == self.MATCH_BY_TARGET else carbon_pair.other(tkn)
            )
            #print(f"[_trade] limit_amt={limit_amt}, tkn={tkn}, match_by={match_by}, limit_price={limit_price}")
            #print(f"[_trade] limit_amt={limit_amt}, tkn={tkn}, rawprice={rawprice}, limit_price={limit_price}")
            
        try:
            decimals = self.decimals
            carbon_pair_r = carbon_pair.reverse

            if not inpair:
                raise NotImplementedError(
                    "Currently only inpair routing implemented", inpair
                )

            # parse applicable orders, indexes, and map the router ids to the simulator ids
            order_ids = [
                k
                for k in self.orders.keys()
                if self.orders[k].pair_name
                   in [carbon_pair.pair_iso, carbon_pair_r.pair_iso]
                   and self.orders[k].tkn == tkn
                   and not self.orders[k].disabled
                if self.orders[k].y > 0
            ]
            use_positions = use_positions if use_positions is not None else order_ids
            applicable_orders = [v for k, v in self.orders.items() if k in order_ids and k in use_positions]
            id_map = {i: applicable_orders[i].id for i in range(len(applicable_orders))}

            if self.debug:
                print("[_trade] order_ids", order_ids)
                print("[_trade] id_map", id_map)
                print("[_trade] applicable_orders", applicable_orders)

            # set the router to only consider the applicable orders
            self.matcher.orders = applicable_orders

            # route the trade
            routes = trade_action(
                x=amt,
                threshold_orders=threshold_orders,
                support_partial=support_partial,
            )

            # retrieve the final route
            final_route = routes[-1]

            #   :tkn:           the token that has been provided to the function
            #   :tkno:          the other token in the pair
            #   :tknb:          the base token of the pair
            #   :tknq:          the quote token of the pair
            tkno = carbon_pair.other(tkn)
            tknb = carbon_pair.tknb
            tknq = carbon_pair.tknq

            # :output:      the asset SOLD by the AMM, and BOUGHT by the TRADER
            #               output is a NEGATIVE number
            # :input:       the asset BOUGHT by the AMM, and SOLD by the TRADER
            #               input is a POSITIVE number
            raw_price = -final_route.total_output / final_route.total_input
            if self.debug:
                print("[_trade] final route", final_route)
                print(
                    f"[_trade] matchmethod={final_route.match_method} tkn={tkn} tkno={tkno}"
                )
                print(
                    f"[_trade] p={raw_price} num={tkn} (amm_in={final_route.total_input} "
                    f"amm_out= {final_route.total_output} mm={final_route.match_method})"
                )

            # the variable `tkn` always holds the token that goes OUT of the AMM
            # in other words: `total_output` is measured in `tkn` and `total_input` in `otkn`
            # therefore, the numeraire of raw_p = out/in is the token associated to out, `tkn`
            price_avg = carbon_pair.convert_price(raw_price, tkn)
            # price_avg = round(float(price_avg), decimals)
            if self.debug:
                print("[_trade] final routes", routes)

            # we are using the function `limit_is_met` from CarbonPair; we need following inputs
            # :tkn:                 the reference token (see below)
            # :limit_price:         the limit price (execute there OR BETTER)*; this is simply the
            #                       `limit_price` parameter from the arguments
            # :buysell:             whether the person who submitted the order (here: the trader)
            #                       BUYs or SELLs (see below)
            # :currrent_price:      the price at which the transaction can be executed*; this is the
            #                        `price_avg` we've just calculated
            # *in units of the pair
            #
            # It is important to understand the following symmetry: a limit BUY order for ETH
            # at 2000 USDC per ETH is the same as a limit SELL order for USDC at 2000. We have
            # seen above that `tkn` is always the token going OUT of the AMM, therefore the
            # trader who placed the limit order BUYs it. Therefore we have
            # - `tkn`[limit_is_met] = `tkn`[function argument], and
            # - `buysell = BUY`

            if limit_price is not None:
                print(f"[_trade] limit_price={limit_price}, price_avg={price_avg}")
                limitfail = not carbon_pair.limit_is_met(
                    tkn, limit_price, carbon_pair.BUY, round(float(price_avg), decimals)
                )
                if limitfail:
                    execute = False
            else:
                limitfail = None

            price_at = []
            for route in routes:
                price_at += [route.input]
                route.input = round(float(route.input), decimals)
                route.output = round(float(route.output), decimals)
                route.total_price = round(abs(float(sum(price_at))), decimals)

            # iterate through the routes
            ct = 0
            trade_ids = []
            numtrades = self.numtrades
            for route in routes:

                # Get the mapped simulator index
                indx = id_map[route.index]
                if self.debug:
                    print("[_trade] final route.index", route.index)
                    print("[_trade] final indx", indx)

                amt1, amt2 = abs(route.output), abs(route.input)

                route_info = trade_description.replace("$", str(amt1)).replace(
                    "*", str(amt2)
                )

                if execute:
                    self.orders[indx].y -= Decimal(amt1) if not (is_by_target and isinstance(self.matcher, FastRouter)) else Decimal(amt2)
                    if self.orders[indx].linked_to_id != indx:
                        self.orders[self.orders[indx].linked_to_id].y += Decimal(amt2) if not (is_by_target and isinstance(self.matcher, FastRouter)) else Decimal(amt1)
                        if (
                                self.orders[self.orders[indx].linked_to_id].y
                                > self.orders[self.orders[indx].linked_to_id].y_int
                        ):
                            # if the linked order is overfilled, update the order y_int
                            self.orders[
                                self.orders[indx].linked_to_id
                            ].y_int = self.orders[self.orders[indx].linked_to_id].y

                # track the route info
                uid = f"{numtrades}.{ct}"
                trade_ids.append(uid)
                self.trades[uid] = {
                    "uid": [uid],  # =id.sub_id
                    "id": [numtrades],  # trade id
                    "subid": [ct],  # id of route withing trade (aggr=A)
                    "note": [f"route #{indx}"],  # human-readable note about this item
                    "aggr": [False],  # True iff this is the aggregate route
                    "exec": [
                        execute
                    ],  # True iff this trade has been executed (also False on limitfail)
                    "partial": [
                        is_partial
                    ],  # True support_partial is set to True
                    "limitfail": [
                        limitfail
                    ],  # False if limit met, True if not, None if no limit
                    "amt1": [
                        amt1 if not (is_by_target and isinstance(self.matcher, FastRouter)) else amt2
                        ],  # the amount of `tkn1` being sold by the AMM (>0)
                    "tkn1": [tkn],  # the token being SOLD by the AMM
                    "amt2": [
                        amt2 if not (is_by_target and isinstance(self.matcher, FastRouter)) else amt1
                        ],  # the amount of `tkn2` being bought by the AMM (>0)
                    "tkn2": [tkno],  #  the token being BOUGHT by the AMM
                    "pair": [
                        carbon_pair.pair_iso
                    ],  # the isoname of the pair, indicating the quote token (eg ETHUSDC)
                    "routeix": [
                        indx
                    ],  # for routes: the index of the route; for aggr: list of routes
                    "nroutes": [
                        1
                    ],  # the number of routes across which the trade was executed
                    "price": [
                        round(float(price_avg), decimals) if not (is_by_target and isinstance(self.matcher, FastRouter)) else round(float(1 / Decimal(price_avg)), decimals)
                    ],  # the price amt_/amt_, in the convention of the pair
                    "p_unit": [
                        f"{tknq} per {tknb}"
                    ],  # the units in which price is quote (eg USD per ETH)

                }

                if isinstance(self.matcher, AlphaRouter):
                    self.trades[uid]["threshold_orders"] = [
                        threshold_orders
                    ]  # the maximum number of order to be routed through using the alpha router

                ct += 1
                if self.debug:
                    print("[_trade] routeinfo", route_info.replace("#", str(indx)))

            # Handle base token vs quote token variable swaps for the trade
            ttl_input = round(abs(routes[-1].total_input), decimals)
            ttl_output = round(abs(routes[-1].total_output), decimals)
            # price_avg = f"{price_avg}"

            amt1, amt2 = ttl_output, ttl_input
            if amt != ttl_input:
                is_partial = True

            note = f"AMM sells {amt1 if not (is_by_target and isinstance(self.matcher, FastRouter)) else amt2:.0f}{tkn} buys {amt2 if not (is_by_target and isinstance(self.matcher, FastRouter)) else amt1:.0f}{tkno}"
            num_trades = len(routes)
            order_ids = [id_map[o.index] for o in routes]
            uid = f"{numtrades}"
            trade_ids.append(uid)
            self.trades[uid] = {
                "uid": [uid],  # see comments above
                "id": [self.numtrades],
                "subid": ["A"],
                "note": [note],
                "aggr": [True],
                "exec": [execute],
                "partial": [is_partial],
                "limitfail": [limitfail],
                "amt1": [
                    amt1 if not (is_by_target and isinstance(self.matcher, FastRouter)) else amt2
                    ],
                "tkn1": [tkn],
                "amt2": [
                    amt2 if not (is_by_target and isinstance(self.matcher, FastRouter)) else amt1
                    ],
                "tkn2": [tkno],
                "pair": [carbon_pair.pair_iso],
                "routeix": [str(order_ids)],
                "nroutes": [num_trades],
                "price": [
                    round(float(price_avg), decimals) if not (is_by_target and isinstance(self.matcher, FastRouter)) else round(float(1 / Decimal(price_avg)), decimals)
                    ],
                "p_unit": [f"{tknq} per {tknb}"],
            }

            if isinstance(self.matcher, AlphaRouter):
                self.trades[uid]["threshold_orders"] = [
                    threshold_orders
                ]  # the maximum number of order to be routed through using the alpha router

            # Get the trade info results
            if len(self.trades) > 0:
                trades = pd.concat([pd.DataFrame(self.trades[i]) for i in self.trades])
                trades = trades[trades["uid"].isin(trade_ids)]

            if self.verbose:
                print(note, f" ({num_trades} routes)")
            self.numtrades += 1

        except Exception as e:
            if self.raiseonerror:
                raise
            return {"success": False, "error": str(e), "exception": e}
        return {"success": True, "trades": trades, "is_by_target": is_by_target, "tkn_sold": tkn}

    def amm_buys(
            self,
            tkn: str,
            amt: Any,
            pair: str = None,
            execute: bool = True,
            inpair: bool = True,
            limit_price: Any = None,
            limit_amt: Any = None,
            threshold_orders: int = 10,
            use_positions: List[int] = None,
            use_positions_matchlevel: List[int] = [],
            support_partial: bool = False,
    ) -> Dict[str, Any]:
        """
        the AMM buys (and the trader sells) `amt` > 0 of `tkn`

        :tkn:               the token bought by the AMM and sold by the trader, eg "ETH"
        :amt:               the amount bought by the AMM and sold by the trader (must be positive)
        :pair:              the token pair to which the trade corresponds, eg "ETHUSD"
        :execute:           if True (default), the trade is executed; otherwise only routing is shown
        :inpair:            if True, only match within pair; if False (default), route through all available pairs
        :limit_price:       the limit price of the order (this price or better from point of view
                            of the trader, not the AMM!), quoted in convention of the pair*
        :limit_amt:         the minimum amount of tokens the traders expects to obtain; quoted in units of the
                            other token, ie `carbon_pair.other(tkn)`*
        :threshold_orders:  the maximum number of order to be routed through using the alpha router
        :use_positions:     the positions to use for the trade (default: all positions)
        :support_partial:   if True (not default), and insufficient liquidity for a trade request, a partial fullfilment is made
        
        *limit_price and limit_amt are redundant; either one of them can be given, or none, but not both
        """

        #print("[amm_buys]", limit_amt, limit_price)
        

        try:

            if amt < 0:
                print(f"[amm_buys] negative amount {amt}; calling amm_sells")
                return self.amm_sells(
                    tkn=tkn,
                    amt=-amt,
                    pair=pair,
                    execute=execute,
                    inpair=inpair,
                    limit_price=limit_price,
                    limit_amt=limit_amt,
                    threshold_orders=threshold_orders,
                    support_partial=support_partial,
                )

            # get the token `tkn`, the other token `tkno` and the CarbonPair object
            tkn, tkno, carbon_pair = self._get_tkn_and_validate(tkn, pair)
            self._assert_position_is_valid(tkno, carbon_pair)
            return self._trade(
                tkn=tkno,
                amt=Decimal(str(amt)),
                carbon_pair=carbon_pair,
                #trade_action=self.matcher.match_by_src,
                match_by = self.MATCH_BY_SOURCE,
                trade_description=f"trader buys $ {tkno} sells * {tkn} | AMM sells $ {tkno} buys * {tkn} via "
                                  f"order #",
                execute=execute,
                limit_price=limit_price,
                limit_amt=limit_amt,
                inpair=inpair,
                threshold_orders=threshold_orders,
                use_positions=use_positions,
                use_positions_matchlevel=use_positions_matchlevel,
                is_by_target=False,
                support_partial=support_partial,
            )
        except Exception as e:
            if self.raiseonerror:
                raise
            return {"success": False, "error": str(e), "exception": e}

    trader_sells = amm_buys

    def amm_sells(
            self,
            tkn: str,
            amt: Any,
            pair: str = None,
            execute: bool = True,
            inpair: bool = True,
            limit_price: Any = None,
            limit_amt: Any = None,
            threshold_orders: int = 10,
            use_positions: List[int] = None,
            use_positions_matchlevel: List[int] = [],
            support_partial: bool = False,
    ) -> Dict[str, Any]:
        """
        the AMM sells (and the trader buys) `amt` > 0 of `tkn`

        :tkn:               the token sold by the AMM and bought by the trader, eg "ETH"
        :amt:               the amount sold by the AMM and bought by the trader (must be positive)
        :pair:              the token pair to which the trade corresponds, eg "ETHUSD"
        :execute:           if True (default), the trade is executed; otherwise only routing is shown
        :inpair:            if True, only match within pair; if False (default), route through all available pairs
        :limit_price:       the limit price of the order (this price or better from point of view
                            of the trader, not the AMM!), quoted in convention of the pair*
        :limit_amt:         the minimum amount of tokens the traders expects to obtain; quoted in units of the
                            other token, ie `carbon_pair.other(tkn)`*
        :threshold_orders:  the maximum number of order to be routed through using the alpha router 
        :use_positions:     the positions to use for the trade (default: all positions)
        :support_partial:   if True (not default), and insufficient liquidity for a trade request, a partial fullfilment is made
        
        *limit_price and limit_amt are redundant; either one of them can be given, or none, but not both
        """

        #print("[amm_sells]", limit_amt, limit_price)
        
        try:

            if amt < 0:
                print(f"[amm_sells] negative amount {amt}; calling amm_buys")
                return self.amm_buys(
                    tkn=tkn,
                    amt=-amt,
                    pair=pair,
                    execute=execute,
                    inpair=inpair,
                    limit_price=limit_price,
                    limit_amt=limit_amt,
                    threshold_orders=threshold_orders,
                    support_partial=support_partial,
                )

            # get the token `tkn`, the other token `tkno` and the CarbonPair object
            tkn, tkno, carbon_pair = self._get_tkn_and_validate(tkn, pair)
            self._assert_position_is_valid(tkn, carbon_pair)

            return self._trade(
                tkn=tkn,
                amt=-Decimal(str(amt)),
                carbon_pair=carbon_pair,
                #trade_action=self.matcher.match_by_target,
                match_by = self.MATCH_BY_TARGET,
                trade_description=f"trader buys $ {tkno} sells * {tkn} | AMM sells $ {tkn} buys * {tkno} via "
                                  f"order #",
                execute=execute,
                limit_price=limit_price,
                limit_amt=limit_amt,
                inpair=inpair,
                threshold_orders=threshold_orders,
                use_positions=use_positions,
                use_positions_matchlevel=use_positions_matchlevel,
                is_by_target=True,
                support_partial=support_partial,

            )

        except Exception as e:
            if self.raiseonerror:
                raise
            return {"success": False, "error": str(e), "exception": e}

    trader_buys = amm_sells

    @staticmethod
    def _to_pandas(order: Order, decimals: int = 50) -> pd.DataFrame:
        """
        Exports Order values for inspection...
        """
        # print("[_to_pandas]", order.pair)

        orderui = CarbonOrderUI.from_order(order)
        # print("[_to_pandas]", orderui)

        if order.disabled:
            p_start = None
            p_end = None
        else:
            p_start = float(order.pair.convert_price(order.p_high, order.tkn))
            p_end   = float(order.pair.convert_price(order.p_low, order.tkn))

        dic = {
            "id":           order.id,
            "pair":         order.pair_name,
            "tkn":          order.tkn,
            # "y_int":      round(order.y_int, decimals),
            # "y":          round(order.y, decimals),
            "y_int":        float(order.y_int),
            "y":            float(order.y),
            "y_unit":       order.tkn,
            # "p_start":    round(order.pair.convert_price(order.p_high, order.tkn), decimals),
            # "p_end":      round(order.pair.convert_price(order.p_low, order.tkn), decimals),
            "disabled":     order.disabled,
            "p_start":      p_start,
            "p_end":        p_end,
            "p_marg":       orderui.p_marg,
            "p_unit":       order.pair.price_convention,
            "lid":          order.linked_to_id,
        }
        return pd.DataFrame(dic, index=[order.id])

    def state(self, pair: str = None, inclhistory: bool = True) -> Dict[str, Any]:
        """
        returns a state record, describing the current state of the system

        :pair:          if None (default), the entire state of the system is provided; if it is set to a specific
                        value, only state corresponding to this specific `pair` is provided
        :inclhistory:   if True (default), also provide history, in particular trade history If False, no trades
                        are provided
        :returns:       a dict with the current state of the system


        RETURN DICT
        :orders:        a pandas data frame showing current orders
        :orderuis:      a list of CarbonOrderUI objects presenting current orders
        :trades:        a pandas data frame showing trades
        """
        try:

            if len(self.orders) > 0:
                if pair is not None:
                    if type(pair) is str:
                        pair = [pair]
                    applicable_orders = [
                        order
                        for order in self.orders.values()
                        if order.pair_name in pair
                    ]
                else:
                    applicable_orders = [order for order in self.orders.values()]
                orders = pd.concat(
                    [
                        self._to_pandas(o, decimals=self.decimals)
                        for o in applicable_orders
                    ]
                )
                orderuis = {o.id: CarbonOrderUI.from_order(o) for o in applicable_orders}
                for o in applicable_orders:
                    orderuis[o.id].set_linked(orderuis[o.linked_to_id])
            else:
                orders = pd.DataFrame({})
                orderuis = dict()

            if len(self.trades) > 0:
                trades = pd.concat([pd.DataFrame(self.trades[i]) for i in self.trades])
            else:
                trades = pd.DataFrame({})

            if self.verbose and False:
                print("[state.orders]")
                print(
                    tabulate(orders, headers=list(orders.columns), tablefmt="pretty"),
                )
                print("\n\n[state.trades]")
                print(
                    tabulate(trades, headers=list(trades.columns), tablefmt="pretty"),
                )
            return {
                "orders": orders,
                "orderuis": orderuis,
                "trades": trades if inclhistory else False,
            }

        except Exception as e:
            if self.raiseonerror:
                raise
            return {"success": False, "error": str(e), "exception": e}

    ASDICT = "asdict"
    ASDF = "asdf"

    def liquidity(self, format=ASDICT):
        """
        returns the aggregate liquidity positions by pair and token, in units of the respective token(!)

        :asdf:      if True, returns the results as dataframe, otherwise as dict

        NOTE: if you pairs are provided in both directions (eg ETHUSDC and USDCETH), which is bad
        practice in any case, then those liquidity positions are not aggregated and the result
        must be corrected manually
        """
        try:
            df = pd.pivot_table(
                self.state()["orders"],
                values=["y"],
                index=["pair", "tkn"],
                aggfunc=np.sum
            )
        except KeyError:
            return None

        if format == self.ASDF:
            return df
        dct = df.to_dict(orient='dict')["y"]  # yields {("ETHUSDC", ETH): ...}
        pairs = set(k[0] for k in dct)  # yields {"ETHUSDC", "ETHDAI", ...}
        result = {p: {k[1]: v for k, v in dct.items() if k[0] == p} for p in
                  pairs}  # yields {"ETHUSDC": {"ETH": ..., "USDC": ...}, ...}
        if format == self.ASDICT:
            return result
        raise ValueError(f"Unknown format {format}")

    @property
    def _posid(self) -> int:
        """
        PRIVATE - auto-incrementing position id
        """
        return next(self._pos_id)

    def _get_tkn_and_validate(self, tkn: str, pair: str) -> Tuple[str, Any, CarbonPair]:
        """
        PRIVATE - handles trade tkn, pair validations
        """
        carbon_pair = self.get_carbon_pair(pair)
        return tkn, carbon_pair.other(tkn), carbon_pair

    def _assert_tkn_is_in_positions(self, tkn: str):
        """
        PRIVATE - checks if tkn has any liquid (=non-empty) positions (raises if not)
        """
        if tkn not in {o.tkn for o in self.orders.values() if o.y > 0}:
            raise ValueError(f"token {tkn} has no non-empty liquidity positions")

    def _assert_pair_is_in_positions(self, carbon_pair: CarbonPair):
        """
        PRIVATE - checks if pair has any liquid positions (raises if not)
        """
        orders = {o.pair_name for o in self.orders.values() if o.y > 0}
        if (
                carbon_pair.pair_iso not in orders
                and carbon_pair.reverse.pair_iso not in orders
        ):
            raise ValueError(f"valid liquidity positions for {carbon_pair.pair_iso}")

    def _assert_position_is_valid(self, tkn: str, carbon_pair: CarbonPair):
        """
        PRIVATE - checks if tkn + pair exist and are valid (raises if not)
        """
        self._assert_tkn_is_in_positions(tkn)
        self._assert_pair_is_in_positions(carbon_pair)

    @property
    def numpos(self) -> int:
        """number of positions"""
        return len(self.orders)

    def __repr__(self):
        pair = f"'{self._carbon_pair.slashpair}'" if self._carbon_pair else None
        return f"{self.__class__.__name__}(<{self.numpos} orders, {self.numtrades} trades>, pair={pair}, mm='{self._mm}', xf={self.exclude_future})"



"""Derived from `passed_tests` notebook test-1"""
from carbon import CarbonSimulatorUI


def test_add_order_concentrated_on_one_point():
    """
    Derived from `passed_tests` notebook test-1
    """
    Sim = CarbonSimulatorUI(pair="USDC/ETH", verbose=True)
    assert Sim.add_order("ETH", 100, 2000, 2000)["orders"].to_dict() == {'id': {0: 0, 1: 1}, 'pair': {0: 'USDCETH', 1: 'USDCETH'}, 'tkn': {0: 'ETH', 1: 'USDC'}, 'y_int': {0: 100.0, 1: 0.0}, 'y': {0: 100.0, 1: 0.0}, 'y_unit': {0: 'ETH', 1: 'USDC'}, 'disabled': {0: False, 1: True}, 'p_start': {0: 2000.0, 1: None}, 'p_end': {0: 2000.0, 1: None}, 'p_marg': {0: 2000.0000000000002, 1: None}, 'p_unit': {0: 'ETH per USDC', 1: 'ETH per USDC'}, 'lid': {0: 1, 1: 0}}


def test_amm_cannot_buy_eth_with_no_usdc():
    """
    Derived from `passed_tests` notebook test-1
    """
    Sim = CarbonSimulatorUI(pair="USDC/ETH", verbose=True)
    Sim.add_order("ETH", 100, 2000, 2000)
    result = Sim.amm_buys("ETH", 2)
    assert result["success"] is False
    assert result["error"] == "token USDC has no non-empty liquidity positions"


def test_amm_can_sell_eth_at_curve_price():
    """
    Derived from `passed_tests` notebook test-1
    """
    Sim = CarbonSimulatorUI(pair="USDC/ETH", verbose=True)
    Sim.add_order("ETH", 100, 2000, 2000)
    result = Sim.amm_sells("ETH", 2)
    assert result["success"] is True
    assert result["trades"].to_dict() == {
        "uid": {0: "0"},
        "id": {0: 0},
        "subid": {0: "A"},
        "note": {0: "AMM sells 2ETH buys 0USDC"},
        "aggr": {0: True},
        "exec": {0: True},
        "limitfail": {0: None},
        "amt1": {0: 2.0},
        "tkn1": {0: "ETH"},
        "amt2": {0: 0.001},
        "tkn2": {0: "USDC"},
        "pair": {0: "USDCETH"},
        "routeix": {0: "[0]"},
        "nroutes": {0: 1},
        "price": {0: 2000.0},
        "p_unit": {0: "ETH per USDC"},
        "partial": {0: True},
    }


def test_can_add_multiple_single_positions():
    """
    Derived from `passed_tests` notebook test-4
    """
    Sim = CarbonSimulatorUI(pair="USDC/ETH", verbose=True)
    test_cases = [
        ("ETH", 0.0005, 2700, 2800),
        ("ETH", 0.0005, 2000, 2700),
        ("USDC", 1000, 2700, 2800),
        ("USDC", 1000, 2000, 2700),
    ]
    for case in test_cases:
        assert Sim.add_order(*case)["success"] is True


def test_no_pair_provided():
    """
    Derived from `passed_tests` notebook test-6
    """
    Sim = CarbonSimulatorUI(verbose=True)
    result = Sim.add_order("ETH", 100, 2000, 2000)
    assert result["success"] is False
    assert (
        result["error"]
        == "Pair must be provided in function or simulation defaults"
    )


def test_geometric_mean_in_one_step():
    """
    Derived from `passed_tests` notebook test-10
    """
    Sim = CarbonSimulatorUI(verbose=True)
    Sim.add_order("ETH", 100, 2000, 3000, "ETH/USDC")
    result = Sim.amm_sells("ETH", 100, "ETH/USDC")
    assert result["success"] is True
    assert round(Sim.state()["trades"]["price"].astype(float).mean(), 4) == 2449.4897


def test_geometric_mean_in_two_steps():
    """
    Derived from `passed_tests` notebook test-10
    """
    Sim = CarbonSimulatorUI(verbose=True)
    Sim.add_order("ETH", 100, 2000, 3000, "ETH/USDC")
    Sim.amm_sells("ETH", 50, "ETH/USDC")
    result = Sim.amm_sells("ETH", 50, "ETH/USDC")
    assert result["success"] is True
    assert round(Sim.state()["trades"]["price"].astype(float).mean(), 4) == 2449.4897



"""Non-exact "fast" router class."""
from decimal import Decimal
from functools import cmp_to_key
from .base_router import *


@dataclass
class FastRouter(BaseRouter):
    """
    This router ranks individual orders by their potential to deliver the best price, and a resultant list of
    “matching objects” is distributed across various orders - sorted by best-to-worst price (and worst-to-best gas
    cost).

    Note: The result of this method is not guaranteed to be optimal, but is typically "close enough" to optimal
    when accounting for gas savings. Additionally, it is simpler and easier to comprehend.
    """

    filter_map = {
        False: {
            'input': 0,
            'output': 0
        },
        True: {
            'input': 0,
            'output': 0
        }
    }

    ONE: DecFloatInt = 2 ** 32
    MIN: DecFloatInt = 0
    MAX: DecFloatInt = 2 ** 128 - 1

    def mul_div_f(self, x: DecFloatInt, y: DecFloatInt, z: DecFloatInt) -> DecFloatInt:
        # #self.logger.debug(f"mul_div_f: {x} {y} {z}")
        if self.use_floor_division:
            return x * y // z
        else:
            return x * y / z

    def mul_div_c(self, x: DecFloatInt, y: DecFloatInt, z: DecFloatInt) -> DecFloatInt:
        # #self.logger.debug(f"mul_div_c: {x} {y} {z}")
        if self.use_floor_division:
            return (x * y + z - 1) // z
        else:
            return x * y / z

    def amt_by_src(
            self, subject: int, dx: DecFloatInt, position_subset: List[int] = None
    ) -> DecFloatInt:
        """
        * Return:
        *
        *                  x * z ^ 2
        * -------------------------------------------
        *  (A * y + B * z) * (A * y + B * z - A * x)
        *
        """
        order = self.orders[subject]
        y, z, A, B = [Decimal(order.y), Decimal(order.y_int), Decimal(order.S), Decimal(order.B)]
        temp1 = z * self.ONE
        temp2 = y * A + z * B
        temp3 = temp2 - dx * A
        res = self.mul_div_c(dx * temp1, temp1, temp2 * temp3)
        # self.logger.debug(
        #     f"amt_by_src: {subject} {dx} -> {res} ({order.y} {order.y_int} {order.S} {order.B})"
        # )
        return res

    def amt_by_target(
            self, subject: int, dx: DecFloatInt, position_subset: List[int] = None
    ) -> DecFloatInt:
        """
        * Return:
        *
        *      x * (A * y + B * z) ^ 2
        * ---------------------------------
        *  A * x * (A * y + B * z) + z ^ 2
        *
        """
        order = self.orders[subject]
        y, z, A, B = Decimal(order.y), Decimal(order.y_int), Decimal(order.S), Decimal(order.B)
        temp1 = y * A + z * B
        temp2 = temp1 * dx // self.ONE
        temp3 = temp2 * A + z * z * self.ONE
        res = self.mul_div_f(temp1, temp2, temp3)
        # self.logger.debug(
        #     f"amt_by_target: {subject} {dx} -> {res} ({order.y} {order.y_int} {order.S} {order.B})"
        # )
        return res

    def get_trade_amounts_by_source_amount(
            self, source_amount: DecFloatInt, subject: int
    ) -> Rate:
        order = self.orders[subject]

        target_amount = self.amt_by_target(dx=source_amount, subject=subject)
        output = min(target_amount, order.y)
        _input = self.amt_by_src(dx=output, subject=subject)
        # self.logger.debug(
        #     f"get_trade_amounts_by_source_amount: {source_amount} -> {output} ({_input})"
        # )
        return Rate(_input, output)

    def get_trade_amounts_by_target_amount(
            self, target_amount: DecFloatInt, subject: int
    ) -> Rate:
        order = self.orders[subject]
        _input = min(target_amount, order.y)
        output = self.amt_by_src(dx=_input, subject=subject)
        # self.logger.debug(
        #     f"get_trade_amounts_by_target_amount: {target_amount} -> {_input} ({output})"
        # )
        return Rate(_input, output)

    @staticmethod
    def cmp_min(x: Quote, y: Quote) -> DecFloatInt:
        lhs = x.rate.output * y.rate.input
        rhs = y.rate.output * x.rate.input
        lt = lhs < rhs
        gt = lhs > rhs
        eq = not lt and not gt
        is_lt = lt or (eq and x.rate.output < y.rate.output)
        is_gt = gt or (eq and x.rate.output > y.rate.output)
        # self.logger.debug(f"cmp_min: {x} {y} -> {is_lt} {is_gt}")
        return is_lt - is_gt

    def cmp_max(self, x: Quote, y: Quote) -> DecFloatInt:
        # self.logger.debug(f"cmp_max: {x} {y}")
        return self.cmp_min(y, x)

    def _filter(self, rate: Rate, is_by_target: bool) -> bool:
        filter_input = self.filter_map[is_by_target]['input']
        filter_output = self.filter_map[is_by_target]['output']
        return rate.input > filter_input and rate.output > filter_output

    def match(
            self,
            x: DecFloatInt,
            is_by_target: bool = True,
            completed_trade: bool = False,
            trade: Callable = None,
            cmp: Callable = None,
            check_sufficient_liquidity: bool = True,
            threshold_orders: int = None,
            support_partial: bool = False,
    ) -> List[Action]:
        """
        * Compute a list of {order index, trade amount} tuples:
        * - Let 'n' denote the initial input amount
        * - Iterate the orders from the highest rate to the lowest rate:
        *   - Let 'm' denote the maximum tradable amount not larger than 'n'
        *   - Add the index of the order along with 'm' to the output matching
        *   - If 'm < n' then subtract 'm' from 'n' and continue, otherwise break
        *
        * Sorting the orders from the highest rate to the lowest rate:
        * - Computing the rate of an order:
        *   - Let 'x' denote the maximum tradable amount not larger than 'n'
        *   - Let 'y' denote the output amount of trading 'x'
        *   - The rate is determined as 'y / x'
        * - Comparing the rates of two orders:
        *   - If the rates are different, then the higher one wins
        *   - If the rates are identical, then the one with a higher value of 'y' wins
        """
        use_positions_matchlevel = []
        # if check_sufficient_liquidity:
        #     self.sufficient_liquidity_exists(x, use_positions_matchlevel)

        actions = []
        x = abs(x)
        for quote in sorted(
                [Quote(n, trade(x, n)) for n in self.indexes], key=cmp_to_key(cmp)
        ):
            if x > quote.rate.input:
                if self._filter(quote.rate, is_by_target):
                    actions.append(
                        Action(
                            index=quote.index,
                            input=quote.rate.input,
                            output=quote.rate.output,
                        )
                    )
                    x -= quote.rate.input
            elif x == quote.rate.input:
                if self._filter(quote.rate, is_by_target):
                    actions.append(
                        Action(
                            index=quote.index,
                            input=quote.rate.input,
                            output=quote.rate.output,
                        )
                    )
                    break
            else:
                if self._filter(quote.rate, is_by_target):
                    actions.append(
                        Action(
                            index=quote.index, input=x, output=trade(x, quote.index).output
                        )
                    )
                    break

        # Format the output
        match_method = "by_src" if not is_by_target else "by_target"
        actions[0].total_output = actions[0].output
        actions[0].total_input = actions[0].input
        for i in range(1, len(actions)):
            actions[i].match_method = match_method
            actions[i].total_output = actions[i - 1].total_output + actions[i].output
            actions[i].total_input = actions[i - 1].total_input + actions[i].input


        for i in range(len(actions)):
            actions[i].total_output = self.frmt(actions[i].total_output)
            actions[i].total_input = self.frmt(actions[i].total_input)
            actions[i].input = self.frmt(actions[i].input)
            actions[i].output = self.frmt(actions[i].output)

        actions[-1].total_output = -actions[-1].total_output

        return actions

    def match_by_src(
            self,
            x: DecFloatInt,
            is_by_target: bool = False,
            check_sufficient_liquidity: bool = True,
            threshold_orders: int = None,
            support_partial: bool = False,

    ) -> List[Action]:
        return self.match(
            x=x,
            trade=self.get_trade_amounts_by_source_amount,
            cmp=self.cmp_min,
            is_by_target=is_by_target,
            completed_trade=False,
            check_sufficient_liquidity=check_sufficient_liquidity,
        )

    def match_by_target(
            self,
            x: DecFloatInt,
            is_by_target: bool = True,
            check_sufficient_liquidity: bool = True,
            threshold_orders: int = None,
            support_partial: bool = False,
    ) -> List[Action]:
        return self.match(
            x=x,
            trade=self.get_trade_amounts_by_target_amount,
            cmp=self.cmp_max,
            is_by_target=is_by_target,
            completed_trade=False,
            check_sufficient_liquidity=check_sufficient_liquidity,
        )

    def __post_init__(self):
        if not self.use_floor_division:
            self.ONE = Decimal("1")
            self.MIN = Decimal("0")
            self.MAX = Decimal(str(2 ** 128 - 1))



# ------------------------------------------------------------
# Auto generated test file `test_050_OrderUI.py`
# ------------------------------------------------------------
# source file   = NBTest_050_OrderUI.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 050
# test comment  = OrderUI
# ------------------------------------------------------------



from carbon import CarbonSimulatorUI, CarbonOrderUI, P, __version__, __date__
from math import sqrt
import numpy as np
from matplotlib import pyplot as plt
print(f"Carbon v{__version__} ({__date__})")
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonSimulatorUI))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonOrderUI))





# ------------------------------------------------------------
# Test      050
# File      test_050_OrderUI.py
# Segment   carbonorderui ids and data
# ------------------------------------------------------------
def test_carbonorderui_ids_and_data():
# ------------------------------------------------------------
    
    Sim = CarbonSimulatorUI(pair="ETH/USDC")
    Sim.add_strategy("ETH", 1, 2000, 2500, 750, 1250, 1000)
    ouis = Sim.state()["orderuis"]
    ouis
    
    
    # +
    # check that the id corresponds to the own dict key
    assert ouis[0].id == 0
    assert ouis[1].id == 1
    
    # check that the lid corresponds to the other dict key
    assert ouis[0].lid == 1
    assert ouis[1].lid == 0
    
    # check that the linked objects are correct
    assert ouis[0].linked is ouis[1]
    assert ouis[1].linked is ouis[0]
    
    # +
    # check curve data
    assert ouis[0].pair.slashpair == "ETH/USDC"
    assert ouis[1].pair.slashpair == "ETH/USDC"
    assert ouis[0].tkn == "ETH"
    assert ouis[1].tkn == "USDC"
    
    # prices
    assert ouis[0].pa == 2000
    assert ouis[0].pb == 2500
    assert ouis[1].pa == 1250
    assert ouis[1].pb == 1000
    
    # raw prices
    assert ouis[0].pa_raw == 1/2000
    assert ouis[0].pb_raw == 1/2500
    assert ouis[1].pa_raw == 1250
    assert ouis[1].pb_raw == 1000
    
    # y, yint, pmarg
    assert ouis[0].y == 1
    assert ouis[0].yint == 1
    assert ouis[0].p_marg == ouis[0].pa
    assert ouis[1].y == 750
    assert ouis[1].yint == 750
    assert ouis[1].p_marg == ouis[1].pa
    # -
    

# ------------------------------------------------------------
# Test      050
# File      test_050_OrderUI.py
# Segment   carbonorderui addliqy
# ------------------------------------------------------------
def test_carbonorderui_addliqy():
# ------------------------------------------------------------
    
    Sim = CarbonSimulatorUI(pair="ETH/USDC")
    Sim.add_strategy("ETH", 1, 2000, 2500, 750, 1250, 1000)
    ouis = Sim.state()["orderuis"]
    ouis
    
    # y, yint, pmarg
    assert ouis[0].y == 1
    assert ouis[0].yint == 1
    assert ouis[0].p_marg == ouis[0].pa
    assert ouis[1].y == 750
    assert ouis[1].yint == 750
    assert ouis[1].p_marg == ouis[1].pa
    
    try:
        ouis[0].addliqy(0.5, expandcurve=False)
        raise RuntimeError("should raise")
    except ValueError as e:
        print(e)
    
    r = ouis[0].addliqy(0.5)
    assert r["y_old"] == 1
    assert r["y"] == 1.5
    assert r["dy"] == 0.5
    assert r["yint_old"] == 1
    assert r["yint"] == 1.5
    assert r["pmarg_old"] == 2000
    assert r["pmarg"] == 2000
    assert r["tkny"] == "ETH"
    assert r["expanded"] == True
    r
    

# ------------------------------------------------------------
# Test      050
# File      test_050_OrderUI.py
# Segment   carbonorderui trading
# ------------------------------------------------------------
def test_carbonorderui_trading():
# ------------------------------------------------------------
    
    Sim = CarbonSimulatorUI(pair="ETH/USDC")
    Sim.add_strategy("ETH", 1, 2000, 2500, 750, 1250, 1000)
    ouis = Sim.state()["orderuis"]
    ouis
    
    # y, yint, pmarg
    assert ouis[0].y == 1
    assert ouis[0].yint == 1
    assert ouis[0].p_marg == ouis[0].pa
    assert ouis[1].y == 750
    assert ouis[1].yint == 750
    assert ouis[1].p_marg == ouis[1].pa
    
    # check that the linked objects are correct
    assert ouis[0].linked is ouis[1]
    assert ouis[1].linked is ouis[0]
    
    # ### trade 1: bysource[selly]
    
    r = ouis[0].selly(0.5)
    l = r["linked"]
    assert r["action"] == "bysource[selly]"
    assert r["executed"] == True
    assert r["y_old"] == 1
    assert r["y"] == 0.5
    assert r["dy"] == 0.5
    assert r["expanded"] == False
    assert round(r["x"],4) == 1055.7281
    assert round(r["dx"],4) == 1055.7281
    assert r["x"] == r["dx"]
    assert r["pmarg_old"] == 2000
    assert round(r["pmarg"],4) == 2229.1236
    assert round(r["dx/dy"],4) == 2111.4562
    assert l["y_old"] == 750
    assert l["y"] == l["y_old"] + l["dy"]
    assert l["dy"] == r["dx"]
    assert l["pmarg_old"] == 1250
    assert l["pmarg"] == 1250
    assert round(l["yint_old"],4) == 750
    assert round(l["yint"],4) == 1805.7281
    assert l["expanded"] == True
    assert l["tkny"] == "USDC"
    l0 = l
    #r
    
    # ### trade 2: bysource[selly]
    
    r = ouis[0].selly(0.1)
    l = r["linked"]
    assert r["action"] == "bysource[selly]"
    assert r["executed"] == True
    assert r["y_old"] == 0.5
    assert r["y"] == 0.4
    assert r["dy"] == 0.1
    assert r["expanded"] == False
    assert round(r["x"],4) == 1281.1529
    assert round(r["dx"],4) == 225.4249
    assert round(r["pmarg_old"],4) == 2229.1236
    assert round(r["pmarg"],4) == 2279.6568
    assert round(r["dx/dy"],4) == 2254.2486
    assert l["y_old"] == l0["y"]
    assert l["y"] == l["y_old"] + l["dy"]
    assert l["dy"] == r["dx"]
    assert l["pmarg_old"] == 1250
    assert l["pmarg"] == 1250
    assert round(l["yint_old"],4) == 1805.7281
    assert round(l["yint"],4) == 2031.1529
    assert l["expanded"] == True
    assert l["tkny"] == "USDC"
    l0 = l
    #r
    
    # ### trade 3: bytarget[buyx]
    
    r = ouis[0].buyx(100)
    l = r["linked"]
    assert r["action"] == "bytarget[buyx]"
    assert r["executed"] == True
    assert r["y_old"] == 0.4
    assert round(r["y"],4) == 0.3563
    assert round(r["dy"],4) == 0.0437
    assert r["expanded"] == False
    assert round(r["x"],4) == 1381.1529
    assert round(r["dx"],4) == 100
    assert round(r["pmarg_old"],4) == 2279.6568
    assert round(r["pmarg"],4) == 2302.2550
    assert round(r["dx/dy"],4) == 2290.9280
    assert l["y_old"] == l0["y"]
    assert l["y"] == l["y_old"] + l["dy"]
    assert l["dy"] == r["dx"]
    assert l["pmarg_old"] == 1250
    assert l["pmarg"] == 1250
    assert round(l["yint_old"],4) == 2031.1529
    assert round(l["yint"],4) == 2131.1529
    assert l["expanded"] == True
    assert l["tkny"] == "USDC"
    l0 = l
    #r
    
    # ### trade 4: byprice[tradeto]
    
    r = ouis[0].tradeto(2330)
    l = r["linked"]
    assert r["action"] == "byprice[tradeto]"
    assert r["executed"] == True
    assert round(r["y_old"],4) == 0.3563
    assert round(r["y"],4) == 0.3036
    assert round(r["dy"],4) == 0.0527
    assert r["expanded"] == False
    assert round(r["x"],4) == 1503.2594
    assert round(r["dx"],4) == 122.1064
    assert round(r["pmarg_old"],4) == 2302.2550
    assert round(r["pmarg"],4) == 2330.0
    assert round(r["dx/dy"],4) == 2316.0860
    assert l["y_old"] == l0["y"]
    assert l["y"] == l["y_old"] + l["dy"]
    assert l["dy"] == r["dx"]
    assert l["pmarg_old"] == 1250
    assert l["pmarg"] == 1250
    assert round(l["yint_old"],4) == 2131.1529
    assert round(l["yint"],4) == 2253.2594
    assert l["expanded"] == True
    assert l["tkny"] == "USDC"
    l0 = l
    #r
    

# ------------------------------------------------------------
# Test      050
# File      test_050_OrderUI.py
# Segment   tradeto basics
# ------------------------------------------------------------
def test_tradeto_basics():
# ------------------------------------------------------------
    
    Sim = CarbonSimulatorUI(pair="ETH/USDC")
    Sim.add_strategy("ETH", 1, 2000, 2500, 750, 1250, 1000)
    ouis = Sim.state()["orderuis"]
    ouis
    
    # y, yint, pmarg
    assert ouis[0].y == 1
    assert ouis[0].yint == 1
    assert ouis[0].p_marg == ouis[0].pa
    assert ouis[1].y == 750
    assert ouis[1].yint == 750
    assert ouis[1].p_marg == ouis[1].pa
    
    # check that the linked objects are correct
    assert ouis[0].linked is ouis[1]
    assert ouis[1].linked is ouis[0]
    
    # +
    # ensure the correct PriceOutOfBoundsError is raised
    try:
        ouis[0].yfromp_f(1000, raiseonerror=True)
        raise RuntimeError("should raise")
    except ouis[0].PriceOutOfBoundsErrorBeyondStart as e:
        print(e)
    
    try:
        ouis[0].yfromp_f(3000, raiseonerror=True)
        raise RuntimeError("should raise")
    except ouis[0].PriceOutOfBoundsErrorBeyondEnd as e:
        print(e)
    
    # +
    # ensure PriceOutOfBoundsError is raised
    try:
        ouis[0].yfromp_f(1000, raiseonerror=True)
        raise RuntimeError("should raise")
    except ouis[0].PriceOutOfBoundsError as e:
        print(e)
        
    try:
        ouis[0].yfromp_f(3000, raiseonerror=True)
        raise RuntimeError("should raise")
    except ouis[0].PriceOutOfBoundsError as e:
        print(e)
    
    # +
    # ensure ValueError is still captured
    try:
        ouis[0].yfromp_f(1000, raiseonerror=True)
        raise RuntimeError("should raise")
    except ValueError as e:
        print(e)
        
    try:
        ouis[0].yfromp_f(3000, raiseonerror=True)
        raise RuntimeError("should raise")
    except ValueError as e:
        print(e)
    # -
    
    # ensure that going the wrong way beyond boundaries leads to dy=0, executed=False
    r = ouis[0].tradeto(1000)
    assert r["action"] == "byprice[tradeto]"
    assert r["dy"] == 0
    assert r["executed"] == False
    #r
    
    # ensure that going the right way yet beyond boundaries leads to full execution
    r = ouis[0].tradeto(3000)
    l = r["linked"]
    assert r["action"] == "byprice[tradeto]"
    assert r["executed"] == True
    assert r["y_old"] == 1
    assert r["y"] == 0
    assert r["dy"] == 1
    assert r["expanded"] == False
    assert round(r["x"],4) == 2236.0680
    assert round(r["dx"],4) == 2236.0680
    assert r["x"] == r["dx"]
    assert r["pmarg_old"] == 2000
    assert r["pmarg"] == 2500
    assert round(r["dx/dy"],4) == 2236.0680
    assert l["y_old"] == 750
    assert l["y"] == l["y_old"] + l["dy"]
    assert l["dy"] == r["dx"]
    assert l["pmarg_old"] == 1250
    assert l["pmarg"] == 1250
    assert round(l["yint_old"],4) == 750
    assert round(l["yint"],4) == 2986.0680
    assert l["expanded"] == True
    assert l["tkny"] == "USDC"
    #r
    
    # ensure you can't trade back
    r = ouis[0].tradeto(2250)
    assert r["action"] == "byprice[tradeto]"
    assert r["dy"] == 0
    assert r["executed"] == False
    #r
    

# ------------------------------------------------------------
# Test      050
# File      test_050_OrderUI.py
# Segment   tradeto over linked orders
# ------------------------------------------------------------
def test_tradeto_over_linked_orders():
# ------------------------------------------------------------
    
    Sim = CarbonSimulatorUI(pair="ETH/USDC")
    Sim.add_strategy("ETH", 1, 2000, 2500, 0, 1250, 1000)
    ouis = Sim.state()["orderuis"]
    ouis
    
    # y, yint, pmarg
    assert ouis[0].y == 1
    assert ouis[0].yint == 1
    assert ouis[0].p_marg == ouis[0].pa
    assert ouis[1].y == 0
    assert ouis[1].yint == 0
    assert ouis[1].p_marg == ouis[1].pa
    
    # check that the linked objects are correct
    assert ouis[0].linked is ouis[1]
    assert ouis[1].linked is ouis[0]
    
    r0 = ouis[0].tradeto(3000)
    r1 = ouis[1].tradeto(3000)
    assert r0["dy"] == 1 
    assert r1["dy"] == 0 
    assert r0["pmarg"] == 2500 
    assert r1["pmarg"] == 1250 
    #r0
    
    r0 = ouis[0].tradeto(500)
    r1 = ouis[1].tradeto(500)
    assert round(r0["dy"],4) == 0
    assert round(r1["dy"],4) == 2236.0680
    assert r0["pmarg"] == 2500 
    assert r1["pmarg"] == 1000 
    #r1
    
    r0 = ouis[0].tradeto(3000)
    r1 = ouis[1].tradeto(3000)
    assert round(r0["dy"],4) == 2
    assert round(r1["dy"],4) == 0
    assert r0["pmarg"] == 2500 
    assert r1["pmarg"] == 1250 
    #r0
    

# ------------------------------------------------------------
# Test      050
# File      test_050_OrderUI.py
# Segment   more tradeto over linked orders
# ------------------------------------------------------------
def test_more_tradeto_over_linked_orders():
# ------------------------------------------------------------
    
    Sim = CarbonSimulatorUI(pair="ETH/USDC")
    Sim.add_strategy("ETH", 1, 2000, 2500, 0, 1250, 1000)
    ouis = Sim.state()["orderuis"]
    ouis
    
    # y, yint, pmarg
    assert ouis[0].y == 1
    assert ouis[0].yint == 1
    assert ouis[0].p_marg == ouis[0].pa
    assert ouis[1].y == 0
    assert ouis[1].yint == 0
    assert ouis[1].p_marg == ouis[1].pa
    
    # check that the linked objects are correct
    assert ouis[0].linked is ouis[1]
    assert ouis[1].linked is ouis[0]
    
    assert ouis[0].tkn == "ETH"
    assert ouis[1].tkn == "USDC"
    ethamt_r  = [ouis[0].y]
    usdcamt_r = [ouis[1].y]
    for ix in range(5):
        
        # trade up to 3000
        r0 = ouis[0].tradeto(3000)
        r1 = ouis[1].tradeto(3000)
        ethamt_r  += [ouis[0].y]
        usdcamt_r += [ouis[1].y]
        
        # trade back down to 500
        r0 = ouis[0].tradeto(500)
        r1 = ouis[1].tradeto(500)
        ethamt_r  += [ouis[0].y]
        usdcamt_r += [ouis[1].y]
    
    #plt.scatter(x=ethamt_r, y=usdcamt_r)
    plt.plot(ethamt_r, usdcamt_r, "-o")
    plt.xlabel("ETH")
    plt.ylabel("USDC")
    plt.grid()
    
    assert [round(x,0) for x in ethamt_r]  == [1, 0, 2, 0, 4, 0, 8, 0, 16, 0, 32]
    assert [round(x,0) for x in usdcamt_r] == [0, 2236, 0, 4472, 0, 8944, 0, 17889, 0, 35777, 0]
    
    


"""
Carbon helper module -- encapsulate paramters for a single strategy
"""
__VERSION__ = "1.0"
__DATE__ = "21/01/2023"

from dataclasses import dataclass


@dataclass
class strategy():
    
    p_buy_a: float
    p_buy_b: float
    p_sell_a: float
    p_sell_b: float 

    amt_rsk: float = 0
    amt_csh: float = 0
        
    rsk: str = "RSK"
    csh: str = "CSH"

    @classmethod
    def from_mwh(cls, m=100, g=0, w=0, amt_rsk=0, amt_csh=0, rsk="RSK", csh="CSH"):
        """
        create class from mid and widths
        
        :m:             the (geometric) middle between the two ranges
        :g:             the geometric gap between the two ranges
        :w:             the geometric width of the range
        :amt_rsk:       the amount of risk asset the strategy is seeded with
        :amt_csh:       the amount of cash the strategy is seeded with
        :rsk:           risk asset name (default RSK)
        :csh:           cash name (default CSH)

            p_buy_a     = m/(1+g)
            p_buy_b     = m/(1+g)/(1+w)
            p_sell_a    = m*(1+g)
            p_sell_b    = m*(1+g)*(1+w)
        """
        p_buy_a     = m/(1+g)
        p_buy_b     = m/(1+g)/(1+w)
        p_sell_a    = m*(1+g)
        p_sell_b    = m*(1+g)*(1+w)

        return cls(p_buy_a, p_buy_b, p_sell_a, p_sell_b, amt_rsk, amt_csh, rsk, csh)
    
    @property
    def descr(s):
        bid_s = f"BID {s.p_buy_b:.1f}-{s.p_buy_a:.1f} [{s.amt_csh} {s.csh}]"
        ask_s = f"ASK {s.p_sell_a:.1f}-{s.p_sell_b:.1f} [{s.amt_rsk} {s.rsk}]"
        return f"{bid_s} -- {ask_s}"
    
    @property
    def p(s):
        """returns a tuple suitable for `add_strategy (*params)`"""
        return (s.rsk, s.amt_rsk, s.p_sell_a, s.p_sell_b, s.amt_csh, s.p_buy_a, s.p_buy_b)
    
    @property
    def slashpair(self):
        return f"{self.rsk}/{self.csh}"

strategy.from_mwh(2000, 0.1, 0.05, amt_rsk=1, rsk="ETH", csh="USD").p


from . import Decimal

Amount = Decimal

class Order:
    def __init__(self, order):
        liq = Decimal(order['liquidity'])
        min = Decimal(order['lowestRate']).sqrt()
        max = Decimal(order['highestRate']).sqrt()
        mid = Decimal(order['marginalRate']).sqrt()
        self.y = liq
        self.z = liq * (max - min) / (mid - min)
        self.A = max - min
        self.B = min
    def __iter__(self):
        y = self.y
        z = self.z
        A = self.A
        B = self.B
        yield 'liquidity'    , y
        yield 'lowestRate'   , B ** 2
        yield 'highestRate'  , (B + A) ** 2
        yield 'marginalRate' , (B + A * y / z) ** 2

def tradeBySourceAmount(x, order):
    y, z, A, B = [order.y, order.z, order.A, order.B]
    n = x * (A * y + B * z) ** 2
    d = A * x * (A * y + B * z) + z ** 2
    return x, n / d

def tradeByTargetAmount(x, order):
    y, z, A, B = [order.y, order.z, order.A, order.B]
    n = x * z ** 2
    d = (A * y + B * z) * (A * y + B * z - A * x)
    return n / d, x



"""Alpha Router restricts the exact algo calculation to specifically n threshold number of orders."""
import itertools
from .exact_router_x0y0n import ExactRouterX0Y0N
from .base_router import *
import pandas as pd
from typing import Any
import numpy as np
from tabulate import tabulate
import heapq


@dataclass
class AlphaRouter(BaseRouter):
    """
    Alpha method - Restricts the exact algo calculation to specifically n threshold number of orders.
    """

    use_positions_matchlevel: List[int] = None
    exact_router: ExactRouterX0Y0N = ExactRouterX0Y0N()

    # returns an enumerated list with only values >= val
    def bigger_than(lst, val):
        return([(i,c) for i,c in lst if c >= val])

    # returns the sum of the values in an enumerated list
    def sum_me(lst):
        return(sum([c for i,c in lst]))

    # returns the first single min value in an enumerated list
    def min_item(lst):
        min_c = min([c for i,c in lst])
        min_c_list = [(i,c) for i,c in lst if c == min_c]
        if len(min_c_list) > 1:
            return(min_c_list[0])
        else:
            return(min_c_list[0])

    # returns the first single max value in an enumerated list
    def max_item(lst):
        max_c = max([c for i,c in lst])
        max_c_list = [(i,c) for i,c in lst if c == max_c]
        if len(max_c_list) > 1:
            return(max_c_list[0])
        else:
            return(max_c_list[0])

    def get_i(lst, wanted_i):
        return([(i,c) for i,c in lst if i == wanted_i][0])

    def sum_list_indexes(lst):
        return(sum([i for i,c in lst]))

    def list_indexes(lst):
        return([i for i,c in lst])

    def assertion_checks(full_fill, current_sum, threshold, max_fill, threshold_list, num_values):
        if full_fill == True:
            assert(current_sum >= threshold)
        else:
            assert(current_sum == max_fill)

        if len(threshold_list) < num_values:
            pass
        else:
            assert(len(set([i for i,c in threshold_list])) == num_values)

    
    def gen_one_order_selector(numbers, threshold, num_values):
        count = 0
        current_sum = 0
        indexed_values = list(enumerate(numbers))
        max_index, max_val = AlphaRouter.max_item(indexed_values)
        threshold_list = indexed_values[:num_values]
        max_fill = sum(sorted(numbers, reverse=True)[:num_values])
        full_fill = False

        # liquidity check to determine partial fill
        if threshold <= max_fill:
            full_fill = True
            # print('Full fulfillment expected')
        else:
            # print(f'Partial fulfillment expected of {max_fill}')
            pass

        # loop over the remaining values
        while indexed_values:
            count += 1
            # set i based on the the value of the first item in the list, offset by the number of values selected
            i = indexed_values[0][0]+num_values-1
            current_sum = AlphaRouter.sum_me(threshold_list)

            # early exit when a partial fill is met
            if (not full_fill) & (current_sum==max_fill):
                # print('1')
                AlphaRouter.assertion_checks(full_fill, current_sum, threshold, max_fill, threshold_list, num_values)
                sum_threshold_list_indexes = AlphaRouter.sum_list_indexes(threshold_list)
                # return(threshold_list, current_sum, threshold, count, sum_threshold_list_indexes)#
                return(AlphaRouter.list_indexes(threshold_list))

            # early exit when we filled the order
            if current_sum >= threshold:
                # print('2')
                AlphaRouter.assertion_checks(full_fill, current_sum, threshold, max_fill, threshold_list, num_values)
                sum_threshold_list_indexes = AlphaRouter.sum_list_indexes(threshold_list)
                # return(threshold_list, current_sum, threshold, count, sum_threshold_list_indexes)#
                return(AlphaRouter.list_indexes(threshold_list))
            else:
                # else iterate over the remaining values dont exceed the the total number of inputs
                if i < len(numbers) - 1:
                    min_index, min_val = AlphaRouter.min_item(threshold_list)
                    next_val = AlphaRouter.get_i(indexed_values, i+1)[1]

                    # if the order hasn't been filled more liquidity is needed so swap out the minimum value and take the next in the list
                    # if the min_val in the threshold list is greater than the max_val in the full list things can't get any better
                    if min_val >= max_val:
                        # print('3')
                        AlphaRouter.assertion_checks(full_fill, current_sum, threshold, max_fill, threshold_list, num_values)
                        sum_threshold_list_indexes = AlphaRouter.sum_list_indexes(threshold_list)
                        # return(threshold_list, current_sum, threshold, count, sum_threshold_list_indexes)#
                        return(AlphaRouter.list_indexes(threshold_list))
                    
                    # however, the next value to insert must be greater than the current value
                    elif next_val > min_val:
                        threshold_list.remove(AlphaRouter.min_item(threshold_list))
                        threshold_list.append(AlphaRouter.get_i(indexed_values, i+1))
                    else:
                        pass
                else:
                    # print('4')
                    AlphaRouter.assertion_checks(full_fill, current_sum, threshold, max_fill, threshold_list, num_values)
                    sum_threshold_list_indexes = AlphaRouter.sum_list_indexes(threshold_list)
                    # return(threshold_list, current_sum, threshold, count, sum_threshold_list_indexes)#
                    return(AlphaRouter.list_indexes(threshold_list))
                indexed_values = indexed_values[1:]

    def get_geoprice(self, subject: int) -> DecFloatInt:
        """
        Determines the geometric price of the subject curve.
        """
        order = self.orders[subject]
        x0 = Decimal(str(order.x0))
        y0 = Decimal(str(order.y0))

        if self.use_floor_division:
            return x0 // y0
        else:
            return x0 / y0

    def amt_by_target(
            self, subject: int, dx: DecFloatInt, position_subset: List[int] = None
    ) -> DecFloatInt:
        """
        An alternative trade by target function that performs a hypothetical
        swap against a subject order not constrained by available liquidity.

        :dx:         the required target amount
        :subject:   the "subject" order index
        """
        order = self.orders[subject]
        S = Decimal(str(order.S))
        B = Decimal(str(order.B))
        y = Decimal(str(order.y))
        y_int = Decimal(str(order.y_int))

        if self.use_floor_division:
            return (
                    dx * y_int ** 2 // ((S * y + B * y_int) * (S * y + B * y_int - S * dx))
            )
        else:
            return (
                    dx * y_int ** 2 / ((S * y + B * y_int) * (S * y + B * y_int - S * dx))
            )

    def match_by_target(
            self,
            x: DecFloatInt,
            is_by_target: bool = True,
            check_sufficient_liquidity: bool = True,
            threshold_orders: int = None,
            support_partial: bool = False,
    ) -> List[Action]:
        """
        The match by target function specfic to the Alpha router that:

        a1.) Run the inputAmount
             through all orders of appropriate pairs and return the hypothetical
             outputAmount for that order. Ignores whether or not there is
             sufficient liquidity associated with these orders.

        a2.) Sort the orders by maximum return (i.e. best price)

        a3.) Take the top n orders which **combined** have available liquidity
             greater than the inputAmount. I.e. cumulatively sum the available
             liquidity from the ranked orders and take the minimum number of
             orders required to fill the trade.

        a4.) Restrict the exact algo calculation to specifically these n orders
             to return the optimal distribution for the full inputAmount
             evaluated within the threshold number of orders.
        """
        if threshold_orders == None:
            threshold_orders = 10

        # (step 1.)

        hypothetical_output_amts = {
            i: self.amt_by_target(dx=x, subject=i) for i in self.indexes
        }

        # (step 2.)
        # Order the amounts
        ordered_amts = {
            j: hypothetical_output_amts[j]
            for j in sorted(
                self.indexes, key=lambda i: hypothetical_output_amts[i], reverse=True
            )
        }

        # (step 3.)
        # Get the available liquidity and return the minimum list of orders with
        # sufficient liquidity to fulfill the inputAmount
        ordered_associated_liquidity = {i:self.orders[i].y for i in ordered_amts.keys()}

        results = pd.DataFrame(
            [
                hypothetical_output_amts.keys(),
                hypothetical_output_amts.values(),
            ],
            index=[
                "indexes",
                "hypothetical_output_amts",
            ],
        )
        results = results.T.copy()
        results.sort_values(
            by=["hypothetical_output_amts", 'indexes'], ascending=[False, True], inplace=True
        )

        results2 = pd.DataFrame(
            [
                ordered_associated_liquidity.keys(),
                ordered_associated_liquidity.values(),
            ],
            index=[
                "ordered_associated_liquidity_keys",
                "ordered_associated_liquidity",
            ],
        )
        results2 = results2.T.copy()

        results = pd.merge(results, results2, how='left', left_on = 'indexes', right_on='ordered_associated_liquidity_keys')

        results.fillna(0, inplace=True)
        results.reset_index(inplace=True, drop=True)
        # print(f'is_by_target {is_by_target}') # True
        # print(tabulate(results,headers=list(results.columns)))
        if support_partial & (results.ordered_associated_liquidity.sum() < abs(x)):
            if x < 0:
                x = -results.ordered_associated_liquidity.sum() + Decimal('0.0000000001')
            else:
                x = results.ordered_associated_liquidity.sum() - Decimal('0.0000000001')
        passed_indexes = AlphaRouter.gen_one_order_selector(results.ordered_associated_liquidity, abs(x), threshold_orders)
        top_n_threshold_orders = [results.indexes[i] for i in passed_indexes]

        # (step 4.)
        # Constrain the exact router to just the top n threshold orders and match
        self.exact_router.orders = self.orders
        use_positions_matchlevel = top_n_threshold_orders  # top_n(threshold_orders)
        return self.exact_router.match(
            x=x,
            is_by_target=is_by_target,
            check_sufficient_liquidity=check_sufficient_liquidity,
            threshold_orders=threshold_orders,
            use_positions_matchlevel=use_positions_matchlevel,
            support_partial=support_partial,
        )

    def amt_by_src(
            self, subject: int, dx: DecFloatInt, position_subset: List[int] = None
    ) -> DecFloatInt:
        """
        An alternative trade by source function that performs a hypothetical
        swap against a subject order not constrained by available liquidity.

        :x:         the required source amount
        :subject:   the "subject" order index
        """
        order = self.orders[subject]
        S = Decimal(str(order.S))
        B = Decimal(str(order.B))
        y = Decimal(str(order.y))
        y_int = Decimal(str(order.y_int))

        if self.use_floor_division:
            return (
                    dx
                    * (S * y + B * y_int) ** 2
                    // (S * dx * (S * y + B * y_int) + y_int ** 2)
            )
        else:
            return (
                    dx
                    * (S * y + B * y_int) ** 2
                    / (S * dx * (S * y + B * y_int) + y_int ** 2)
            )

    def match_by_src(
            self,
            x: DecFloatInt,
            is_by_target: bool = False,
            check_sufficient_liquidity: bool = True,
            threshold_orders: int = None,
            support_partial: bool = False,
    ) -> List[Action]:
        """
        The match by target function specfic to the Alpha router that:

        a1.) Run the inputAmount
             through all orders of appropriate pairs and return the hypothetical
             outputAmount for that order. Ignores whether or not there is
             sufficient liquidity associated with these orders.

        a2.) Sort the orders by maximum return (i.e. best price)

        a3.) Take the top n orders which **combined** have available liquidity
             greater than the inputAmount. I.e. cumulatively sum the available
             liquidity from the ranked orders and take the minimum number of
             orders required to fill the trade.

        a4.) Restrict the exact algo calculation to specifically these n orders
             to return the optimal distribution for the full inputAmount
             evaluated within the threshold number of orders.
        """
        # (step 1.)
        hypothetical_output_amts = {i: self.amt_by_src(dx=x, subject=i) for i in self.indexes}

        # (step 3.)
        # Get the available liquidity and return the minimum list of orders with
        # sufficient liquidity to fulfill the inputAmount
        associated_liquidity = [self.orders[i].y for i in self.indexes]

        amounts = []
        effective_prices = []
        available_value = []
        for k, v in hypothetical_output_amts.items():
            if v > self.orders[k].y:
                price = self.get_geoprice(k)
                amount = self.orders[k].y
                amounts += [amount]
                effective_prices += [price]
                available_value += [price * amount]
            else:
                amounts += [v]
                price = x / v
                effective_prices += [price]
                available_value += [price * v]

        results = pd.DataFrame(
            [
                hypothetical_output_amts.keys(),
                hypothetical_output_amts.values(),
                associated_liquidity,
                amounts,
                effective_prices,
                available_value,
            ],
            index=[
                "indexes",
                "hypothetical_output_amts",
                "associated_liquidity",
                "amount",
                "effective_prices",
                "available_value",
            ],
        )
        results = results.T.copy()
        results.sort_values(
            by=["hypothetical_output_amts", 'indexes'], ascending=[False, True], inplace=True
        )

        results.fillna(0, inplace=True)
        results.reset_index(inplace=True, drop=True)
        # print(f'is_by_target {is_by_target}') #False
        # print(tabulate(results,headers=list(results.columns)))
        if support_partial & (results.available_value.sum() < abs(x)):
            if x < 0:
                x = -results.available_value.sum() + Decimal('0.0000000001')
            else:
                x = results.available_value.sum() - Decimal('0.0000000001')
        passed_indexes = AlphaRouter.gen_one_order_selector(results.available_value, abs(x), threshold_orders)
        top_n_threshold_orders = [results.indexes[i] for i in passed_indexes]

        # (step 4.)
        # Constrain the exact router to just the top n threshold orders and match
        self.exact_router.orders = self.orders
        self.use_positions_matchlevel = top_n_threshold_orders  # top_n(threshold_orders)
        return self.match(
            x=x,
            is_by_target=is_by_target,
            check_sufficient_liquidity=check_sufficient_liquidity,
            threshold_orders=threshold_orders,
        )

    def match(
            self,
            x: DecFloatInt,
            is_by_target: bool = True,
            completed_trade: bool = False,
            trade: Callable = None,
            cmp: Callable = None,
            check_sufficient_liquidity: bool = True,
            threshold_orders: int = None,
    ) -> List[Action]:
        """
        Main algorithm to handle matching a trade amount against the curves/orders.
        """
        return self.exact_router.match(
            x=x,
            is_by_target=is_by_target,
            check_sufficient_liquidity=check_sufficient_liquidity,
            threshold_orders=threshold_orders,
            use_positions_matchlevel=self.use_positions_matchlevel,
        )



# ------------------------------------------------------------
# Auto generated test file `test_040_PairInSim.py`
# ------------------------------------------------------------
# source file   = NBTest_040_PairInSim.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 040
# test comment  = PairInSim
# ------------------------------------------------------------



from carbon import CarbonSimulatorUI, P, __version__, __date__
print(f"Carbon v{__version__} ({__date__})")
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonSimulatorUI))




ETHUSDC = P("ETH/USDC")


# ------------------------------------------------------------
# Test      040
# File      test_040_PairInSim.py
# Segment   Pair without CarbonPair defaults
# ------------------------------------------------------------
def test_pair_without_carbonpair_defaults():
# ------------------------------------------------------------
    
    Sim = CarbonSimulatorUI(raiseonerror=True)
    assert(Sim.slashpair is None)
    Sim
    
    try:
        Sim.add_order("ETH", 10, 2000, 3000)
    except ValueError as e:
        print(e)
        assert(str(e)=="Pair must be provided in function or simulation defaults")
    
    try:
        Sim.add_order("ETH", 10, 2000, 3000, pair="ETHUSDC")["orders"]
    except ValueError as e:
        print(e)
        assert(str(e)=="('Illegal slashpair', 'ETHUSDC')")
    
    try:
        Sim.add_order("LINK", 10, 2000, 3000, pair=ETHUSDC)
    except ValueError as e:
        print(e)
        assert str(e) == "('Token not in pair', 'LINK', 'ETH/USDC', 'ETH/USDC')"
    
    o = Sim.add_order("ETH", 10, 2000, 3000, pair=ETHUSDC)["orders"]
    assert(o.iloc[0]["pair"]=="ETHUSDC")
    assert(o.iloc[0]["tkn"]=="ETH")
    assert(o.iloc[0]["p_unit"]=="USDC per ETH")
    o
    
    Sim.add_order("ETH", 10, 2000, 3000, pair="ETH/USDC")["orders"]
    
    try:
        Sim.add_strategy("ETH", 10, 2000, 3000, 10000, 1000, 750)["orders"]
    except ValueError as e:
        print(e)
        assert(str(e)=="Pair must be provided in function or simulation defaults")
    
    Sim.add_strategy("ETH", 10, 2000, 3000, 10000, 1000, 750, pair="ETH/USDC")["orders"]
    

# ------------------------------------------------------------
# Test      040
# File      test_040_PairInSim.py
# Segment   Pair with CarbonPair defaults
# ------------------------------------------------------------
def test_pair_with_carbonpair_defaults():
# ------------------------------------------------------------
    
    SimP = CarbonSimulatorUI(pair="ETH/USDC", raiseonerror=True)
    assert(SimP.slashpair=="ETH/USDC")
    SimP
    
    o = SimP.add_order("ETH", 10, 2000, 3000)["orders"]
    assert(o.iloc[0]["pair"]=="ETHUSDC")
    assert(o.iloc[0]["tkn"]=="ETH")
    assert(o.iloc[0]["p_unit"]=="USDC per ETH")
    o
    
    o = SimP.add_order("ETH", 10, 2000, 3000, pair=ETHUSDC)["orders"]
    assert(o.iloc[0]["pair"]=="ETHUSDC")
    assert(o.iloc[0]["tkn"]=="ETH")
    assert(o.iloc[0]["p_unit"]=="USDC per ETH")
    o
    
    SimP.add_order("ETH", 10, 2000, 3000, pair="ETH/USDC")["orders"]
    
    o = SimP.add_order("ETH", 10, 2000, 3000, pair="ETH/DAI")["orders"]
    assert(o.iloc[0]["pair"]=="ETHDAI")
    assert(o.iloc[0]["tkn"]=="ETH")
    assert(o.iloc[0]["p_unit"]=="DAI per ETH")
    o
    
    SimP.add_strategy("ETH", 10, 2000, 3000, 10000, 1000, 750)["orders"]
    
    
    